# useful functions
def test_attr(attr, typ, name):
    """
    This function tests if the attribute "attr" belongs to the type "typ",
    if not it gives an error message informing the name of the variable
    """
    try:
        return typ(attr)
    except:
        raise TypeError('"{}" keyword must be a {} object'.format(name, typ))




from .star import Star
from .ephem import EphemKernel, EphemJPL, EphemPlanete
from sora.config import input_tests
import astropy.units as u
import astropy.constants as const
from astropy.coordinates import SkyCoord, EarthLocation, Angle, get_sun
from astropy.coordinates import get_moon, GCRS, ITRS, SkyOffsetFrame
from astropy.time import Time
from astropy.table import Table, Row, Column
from astroquery.vizier import Vizier
import numpy as np
import warnings
import os
import matplotlib.pyplot as plt
import glob


class PredictRow(Row):
    """ An Astropy Row object modified for Prediction purposes.
    """
    def plot_occ_map(self, **kwargs):
        """
        Parameters:
            radius (int,float): The radius of the shadow. If not given it uses saved value

            nameimg (str): Change the name of the image saved.
            path (str): Path to a directory where to save map.
            resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
                "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
            states (bool): True to plot the state division of the countries. The states of
                some countries will only be shown depending on the resolution.
            zoom (int, float): Zooms in or out of the map.
            centermap_geo (list): Center the map given coordinates in longitude and latitude.
                It must be a list with two numbers. Default=None.
            centermap_delta (list): Displace the center of the map given displacement
                in X and Y, in km. It must be a list with two numbers. Default=None.
            centerproj (list): Rotates the Earth to show occultation with the center
                projected at a given longitude and latitude. It must be a list with two numbers
            labels (bool): Plots text above and below the map with the occultation parameters.
                Default=True.
            meridians (int): Plots lines representing the meridians for given interval. Default=30 deg
            parallels (int): Plots lines representing the parallels for given interval. Default=30 deg
            sites (dict): Plots site positions in map. It must be a python dictionary where the key is
                the name of the site, and the value is a list with longitude, latitude, delta_x,
                delta_y and color. delta_x and delta_y are displacement, in km, from the point
                of the site in the map and the name. color is the color of the point.
            site_name (bool): If True, it prints the name of the sites given, else it plots only the points
            countries (dict): Plots the names of countries. It must be a python dictionary where the key
                is the name of the country and the value is a list with longitude and latitude
                of the lower left part of the text.
            offset (list): applies an offset to the ephemeris, calculating new CA and instant of CA.
                It is a pair of delta_RA*cosDEC and delta_DEC.
            mapstyle (int): Define the color style of the map. 1 is the default black and white scale.
                2 is a colored map.
            error (int,float): Ephemeris error in mas. It plots a dashed line representing radius + error.
            ercolor (str): Changes the color of the lines of the error bar.
            ring (int,float): It plots a dashed line representing the location of a ring.
                It is given in km, from the center.
            rncolor (str): Changes the color of ring lines.
            atm (int,float): plots a dashed line representing the location of an atmosphere.
                It is given in km, from the center.
            atcolor (str): Changes the color of atm lines.
            chord_delta (list): list with distances from center to plot chords
            chord_geo (2d-list): list with pairs of coordinates to plot chords
            chcolor (str): color of the line of the chords. Default: grey
            heights (list): plots a circular dashed line showing the locations where the observer
                would observe the occultation at a given height above the horizons.
                This must be a list.
            hcolor (str): Changes the color of the height lines.
            mapsize (list): The size of figure, in cm. It must be a list with two values.
                Default = [46.0, 38.0].
            cpoints (int,float): Interval for the small points marking the center of shadow,
                in seconds. Default=60.
            ptcolor (str): Change the color of the center points.
            alpha (float): The transparency of the night shade, where 0.0 is full transparency
                and 1.0 is full black. Default = 0.2.
            fmt (str): The format to save the image. It is parsed directly by matplotlib.pyplot.
                Default = 'png'
            dpi (int): "Dots per inch". It defines the quality of the image. Default = 100.
            lncolor (str): Changes the color of the line that represents the limits of the shadow over Earth.
            outcolor (str): Changes the color of the lines that represents the limits of the shadow outside Earth
            nscale (int,float): Arbitrary scale for the size of the name of the site.
            cscale (int,float): Arbitrary scale for the name of the country.
            sscale (int,float): Arbitrary scale for the size of point of the site.
            pscale (int,float): Arbitrary scale for the size of the points that represent the center of the shadow
            arrow (bool): If true, it plots the arrow with the occultation direction.

            Comment: Only one of centermap_geo and centermap_delta can be given
        """
        radius = kwargs.pop('radius', self.meta['radius'])
        plot_occ_map(self.meta['name'], radius, coord=self['ICRS Star Coord at Epoch'], time=self['Epoch'],
                     ca=float(self['C/A']), pa=float(self['P/A']), vel=float(self['Vel']), dist=float(self['Dist']),
                     mag=float(self['G*']), longi=float(self['long']), **kwargs)


class PredictionTable(Table):
    """ An Astropy Table object modified for Prediction purposes.
    """
    Row = PredictRow

    def __init__(self, *args, **kwargs):
        if 'time' in kwargs:
            values = {}
            time = Time(kwargs['time'])
            time.delta_ut1_utc = 0.0
            time.format = 'iso'
            if time.isscalar:
                time = Time([time])
            values['Epoch'] = Column(time)

            def coord_fmt(value):
                return value.to_string('hmsdms', precision=5, sep=' ')

            coord = SkyCoord(kwargs['coord_star'], unit=(u.hourangle, u.deg))
            values['ICRS Star Coord at Epoch'] = Column(coord, format=coord_fmt)
            try:
                coord_geo = SkyCoord(kwargs['coord_obj'])
            except:
                coord_geo = SkyCoord(kwargs['coord_obj'], unit=(u.hourangle, u.deg))
            values['Geocentric Object Position'] = Column(coord_geo, format=coord_fmt)
            values['C/A'] = Column(kwargs['ca'], format='5.3f', unit='arcsec')
            values['P/A'] = Column(kwargs['pa'], format='6.2f', unit='deg')
            values['Vel'] = Column(kwargs['vel'], format='-6.2f', unit='km/s')
            values['Dist'] = Column(kwargs['dist'], format='7.3f', unit='AU')
            del(kwargs['time'], kwargs['coord_star'], kwargs['coord_obj'], kwargs['ca'],
                kwargs['pa'], kwargs['vel'], kwargs['dist'])
            if 'mag' not in kwargs.keys() and 'mag_20' not in kwargs.keys():
                raise ValueError('User must provide "mag" or "mag_20" parameters')
            if 'mag' in kwargs.keys():
                values['G'] = Column(kwargs['mag'], format='6.3f', unit='mag')
                del kwargs['mag']
            else:
                values['G'] = kwargs['mag_20'] - 2.5*np.log10(np.absolute(values['Vel'])/20.0)
                values['G'].unit = 'mag'
                values['G'].format = '6.3f'
            if 'mag_20' in kwargs.keys():
                values['G*'] = Column(kwargs['mag_20'], format='6.3f', unit='mag')
                del kwargs['mag_20']
            else:
                values['G*'] = values['G'] + 2.5*np.log10(np.absolute(values['Vel'])/20.0)
                values['G*'].unit = 'mag'
                values['G*'].format = '6.3f'
            if 'long' in kwargs.keys():
                values['long'] = Column(kwargs['long'], unit='deg', format='3.0f')
                del kwargs['long']
            else:
                values['long'] = Column((coord.ra - time.sidereal_time('mean', 'greenwich')).wrap_at(360*u.deg),
                                        unit='deg', format='3.0f')
            if 'loct' in kwargs.keys():
                values['loct'] = Column(kwargs['loct'], unit='hh:mm')
                del kwargs['loct']
            else:
                longi = (coord.ra - time.sidereal_time('mean', 'greenwich')).wrap_at(360*u.deg)
                ntime = time + longi.hour*u.hour
                values['loct'] = Column(['{}'.format(t.iso[11:16]) for t in ntime], unit='hh:mm')
            moon_pos = get_moon(time)
            values['M-G-T'] = Column(moon_pos.separation(coord), unit='deg', format='3.0f')
            sun_pos = get_sun(time)
            values['S-G-T'] = Column(sun_pos.separation(coord), unit='deg', format='3.0f')
            if 'source' in kwargs.keys():
                values['GAIA-DR2 Source ID'] = Column(kwargs['source'])
                del kwargs['source']
            else:
                values['GAIA-DR2 Source ID'] = Column(np.repeat('', len(time)))
            super().__init__(values, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    def __itens_by_epoch(self, date):
        """ Gets item list for all occultations that matches the given date

        Parameters:
            date (str): date to match

        Returns:
            item (list): the list of occultations that matches the date
        """
        col = self['Epoch']
        arr = list([i for i, c in enumerate(col) if c.iso.startswith(date)])
        return arr

    def __getitem__(self, item):
        """ The redefinition of __getitem__ allows for selecting prediction based on the ISO date of the event
        """
        if isinstance(item, str) and item not in self.colnames:
            arr = self.__itens_by_epoch(item)
            if len(arr) == 0:
                raise KeyError('No prediction corresponds to time "{}"'.format(item))
            elif len(arr) > 1:
                return self[arr]
            else:
                return self.Row(self, arr[0])
        return super().__getitem__(item)

    @classmethod
    def from_praia(cls, filename, name, **kwargs):
        """ Creates a PredictionTable Table reading from a PRAIA table

        Parameters:
            filename (str): path to the PRAIA table file.
            name (str): Name of the Object of the prediction.
            radius (int,float): Object radius, in km. (not required)
                If not given it's searched in online database.
                If not found online, the default is set to zero.

        Returns:
            A PredictionTable
        """
        from .ephem import read_obj_data
        if not os.path.isfile(filename):
            raise IOError('File {} not found'.format(filename))
        input_tests.check_kwargs(kwargs, allowed_kwargs=['radius'])
        try:
            dados = np.loadtxt(
                filename, skiprows=41,
                usecols=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 25, 26, 27, 28, 29),
                dtype={'names': ('dia', 'mes', 'ano', 'hor', 'min', 'sec', 'afh', 'afm', 'afs', 'ded',
                                 'dem', 'des', 'afh2', 'afm2', 'afs2', 'ded2', 'dem2', 'des2', 'ca',
                                 'pa', 'vel', 'delta', 'mR', 'mK', 'long', 'loct', 'ora', 'ode'),
                       'formats': ('S30', 'S30', 'S30', 'S30', 'S30', 'S30', 'S20', 'S20', 'S20',
                                   'S20', 'S20', 'S20', 'S20', 'S20', 'S20', 'S20', 'S20', 'S20',
                                   'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'S20', 'f4', 'f4')}, ndmin=1)
        except:
            raise IOError('{} is not in PRAIA format or does not have any occultation'.format(filename))

        f = open(filename, 'r')
        lines = f.readlines()
        f.close()

        max_ca = float(lines[14].split()[-2])*u.arcsec

        # reading coordinates
        coor = np.char.array(dados['afh'], unicode=True)
        for i in ['afm', 'afs', 'ded', 'dem', 'des']:
            coor = np.core.defchararray.add(coor, ' ')
            coor = np.core.defchararray.add(coor, np.char.array(dados[i], unicode=True))
        coord_star = SkyCoord(coor, frame='icrs', unit=(u.hourangle, u.degree))

        coor = np.char.array(dados['afh2'], unicode=True)
        for i in ['afm2', 'afs2', 'ded2', 'dem2', 'des2']:
            coor = np.core.defchararray.add(coor, ' ')
            coor = np.core.defchararray.add(coor, np.char.array(dados[i], unicode=True))
        coord_obj = SkyCoord(coor, frame='icrs', unit=(u.hourangle, u.degree))

        # reading time
        tim = np.char.array(dados['ano'], unicode=True)
        len_iso = ['-', '-', ' ', ':', ':']
        arr = ['mes', 'dia', 'hor', 'min', 'sec']
        for i in np.arange(len(arr)):
            tim = np.core.defchararray.add(tim, len_iso[i])
            tim = np.core.defchararray.add(tim, np.char.array(dados[arr[i]], unicode=True))
        time = Time(np.char.array(tim) + '000')

        # defining parameters
        data = read_obj_data()
        radius, error_ra, error_dec = data.get(name.lower(), [0, 0, 0])
        radius = kwargs.get('radius', radius)*u.km
        meta = {'name': name, 'radius': radius, 'max_ca': max_ca, 'ephem': lines[17].split()[-1],
                'error_ra': error_ra*1000, 'error_dec': error_dec*1000}
        return cls(time=time, coord_star=coord_star, coord_obj=coord_obj, ca=dados['ca'], pa=dados['pa'],
                   vel=dados['vel'], mag_20=dados['mR'], dist=dados['delta'],
                   long=dados['long'], loct=dados['loct'], meta=meta)

    def to_praia(self, filename):
        """ Writes PredictionTable to PRAIA format.

        Parameters:
            filename(str): name of the file to save table
        """
        from sora.config.variables import praia_occ_head
        f = open(filename, 'w')
        f.write(praia_occ_head.format(max_ca=self.meta['max_ca'].to(u.arcsec), size=len(self),
                                      ephem=self.meta.get('ephem', 'ephem')))
        for time, coord, coord_geo, ca, pa, vel, dist, mag, mag_20, longi, loct, md, sd, source in self.iterrows():
            dmag = mag_20-mag
            f.write("\n {} {} {}  {}  {}   {}   {:5.3f}  {:6.2f} {:-6.2f} {:5.2f} {:4.1f} "
                    "{:-4.1f} {:-4.1f} {:-4.1f}   {:3.0f}. {}       0.0      0.0 ok g2 0    0    0    0    0".
                    format(time.iso[8:10], time.iso[5:7], time.iso[:4], time.iso[11:21].replace(':', ' '),
                           coord.to_string('hmsdms', precision=4, sep=' '), coord_geo.to_string('hmsdms', precision=4, sep=' '),
                           ca, pa, vel, dist, mag_20, dmag, dmag, dmag, longi, loct))
        f.close()

    def to_ow(self, ow_des, mode='append'):
        """ Writes PredictionTable to OccultWatcher feeder update file format.
            Tables will be saved in two files: "tableOccult_update.txt" and "LOG.dat"

        Parameters:
            ow_des (str): Occult Watcher designation for the object.
            mode (str): 'append' to append table to already existing file, default
                'restart' to overwrite existing file.
        """
        from sora.config.variables import ow_occ_head
        modes = {'append': 'a+', 'restart': 'w'}
        if mode not in modes.keys():
            raise ValueError('mode param must be "append" or "restart".')

        f = open('tableOccult_update.txt', modes[mode])
        if self.meta['radius'] == 0.0:
            warnings.warn('radius is 0.0, please update the value manually in "tableOccult_update.txt" '
                          'before submitting the file')
        if (self.meta['error_ra'] == 0.0) or (self.meta['error_ra'] == 0.0):
            warnings.warn('error_ra and/or error_dec is 0.0, please update the value manually in '
                          '"tableOccult_update.txt" before submitting the file')
        f.write(ow_occ_head.format(name=self.meta['name'], ephem=self.meta.get('ephem', 'ephem'),
                                   max_ca=self.meta['max_ca'].to(u.arcsec), size=len(self),
                                   radius=self.meta['radius'], ow_des=ow_des))
        for time, coord, coord_geo, ca, pa, vel, dist, mag, mag_20, longi, loct, md, sd, source in self.iterrows():
            dmag = mag_20-mag
            f.write('{} {} {}  {}   {}   {}   {:5.3f}  {:6.2f} {:-7.3f} {:7.3f} '
                    '{:4.1f} {:-4.1f}   {:3.0f}. {}  {:4.0f}  {:4.0f}\n'.
                    format(time.iso[8:10], time.iso[5:7], time.iso[:4], time.iso[11:20].replace(':', ' '),
                           coord.to_string('hmsdms', precision=4, sep=' ')[:-1],
                           coord_geo.to_string('hmsdms', precision=4, sep=' ')[:-1],
                           ca, pa, vel, dist, mag_20, dmag, longi, loct,
                           self.meta.get('error_ra', 0), self.meta.get('error_dec', 0))
                    )
        f.write(' '+'-'*148+'\n')
        f.close()

        f = open('LOG.dat', modes[mode])
        t = Time.now()
        for time in self['Epoch']:
            t0 = Time(time.isot.split('T')[0] + ' 00:00:00.0')
            dt = (time-t0).jd*24
            f.write('{} {:5s} {}-{:06.3f}\n'.format(t.isot[:-7], ow_des, time.isot.split('T')[0], dt))
        f.close()

    def plot_occ_map(self, **kwargs):
        """
        Parameters:
            radius (int,float): The radius of the shadow. If not given it uses saved value

            nameimg (str): Change the name of the image saved.
            path (str): Path to a directory where to save map.
            resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
                "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
            states (bool): True to plot the state division of the countries. The states of
                some countries will only be shown depending on the resolution.
            zoom (int, float): Zooms in or out of the map.
            centermap_geo (list): Center the map given coordinates in longitude and latitude.
                It must be a list with two numbers. Default=None.
            centermap_delta (list): Displace the center of the map given displacement
                in X and Y, in km. It must be a list with two numbers. Default=None.
            centerproj (list): Rotates the Earth to show occultation with the center
                projected at a given longitude and latitude. It must be a list with two numbers
            labels (bool): Plots text above and below the map with the occultation parameters.
                Default=True.
            meridians (int): Plots lines representing the meridians for given interval. Default=30 deg
            parallels (int): Plots lines representing the parallels for given interval. Default=30 deg
            sites (dict): Plots site positions in map. It must be a python dictionary where the key is
                the name of the site, and the value is a list with longitude, latitude, delta_x,
                delta_y and color. delta_x and delta_y are displacement, in km, from the point
                of the site in the map and the name. color is the color of the point.
            site_name (bool): If True, it prints the name of the sites given, else it plots only the points
            countries (dict): Plots the names of countries. It must be a python dictionary where the key
                is the name of the country and the value is a list with longitude and latitude
                of the lower left part of the text.
            offset (list): applies an offset to the ephemeris, calculating new CA and instant of CA.
                It is a pair of delta_RA*cosDEC and delta_DEC.
            mapstyle (int): Define the color style of the map. 1 is the default black and white scale.
                2 is a colored map.
            error (int,float): Ephemeris error in mas. It plots a dashed line representing radius + error.
            ercolor (str): Changes the color of the lines of the error bar.
            ring (int,float): plots a dashed line representing the location of a ring.
                It is given in km, from the center.
            rncolor (str): Changes the color of ring lines.
            atm (int,float): plots a dashed line representing the location of an atmosphere.
                It is given in km, from the center.
            atcolor (str): Changes the color of atm lines.
            chord_delta (list): list with distances from center to plot chords
            chord_geo (2d-list): list with pairs of coordinates to plot chords
            chcolor (str): color of the line of the chords. Default: grey
            heights (list): It plots a circular dashed line showing the locations where the observer
                would observe the occultation at a given height above the horizons.
                This must be a list.
            hcolor (str): Changes the color of the height lines.
            mapsize (list): The size of figure, in cm. It must be a list with two values.
                Default = [46.0, 38.0].
            cpoints (int,float): Interval for the small points marking the center of shadow,
                in seconds. Default=60.
            ptcolor (str): Change the color of the center points.
            alpha (float): The transparency of the night shade, where 0.0 is full transparency
                and 1.0 is full black. Default = 0.2.
            fmt (str): The format to save the image. It is parsed directly by matplotlib.pyplot.
                Default = 'png'
            dpi (int): "Dots per inch". It defines the quality of the image. Default = 100.
            lncolor (str): Changes the color of the line that represents the limits of the shadow over Earth.
            outcolor (str): Changes the color of the lines that represents the limits of the shadow outside Earth
            nscale (int,float): Arbitrary scale for the size of the name of the site.
            cscale (int,float): Arbitrary scale for the name of the country.
            sscale (int,float): Arbitrary scale for the size of point of the site.
            pscale (int,float): Arbitrary scale for the size of the points that represent the center of the shadow
            arrow (bool): If true, it plots the arrow with the occultation direction.

            Comment: Only one of centermap_geo and centermap_delta can be given
        """
        for i in range(len(self)):
            self[i].plot_occ_map(**kwargs)

    def remove_occ(self, date):
        """ Removes stellar occultations from table

        Parameters:
            date (str,list): Date or list of dates of the occultation to be removed.
                The dates mut be as shown in the 'Epoch' column. If the date is not
                complete, the function will select all occultations that matches the given string.
                For instance, date='2020-06' will remove all occultations from the month of June 2020.
        """
        if type(date) == str:
            date = [date]
        itens = list(set([y for d in date for y in self.__itens_by_epoch(d)]))
        self.remove_rows(itens)

    def keep_from_selected_images(self, path='.'):
        """ Keeps predictions which images were not deleted in given path
            This function uses the name of the images to identify predictions.
            The name must be the automatic one generated by plot_occ_map().
            The format of the image is not relevant.

        Parameters:
            path (str): path where images are located
        """
        itens = []
        for i, tca in enumerate(self['Epoch']):
            name = '{}_{}'.format(self.meta['name'], tca.isot)
            if not glob.glob(os.path.join(path, name)+'*'):
                itens.append(i)
        self.remove_rows(itens)


def occ_params(star, ephem, time):
    """ Calculates the parameters of the occultation, as instant, CA, PA.

    Parameters:
        star (Star): The coordinate of the star in the same reference frame as the ephemeris.
            It must be a Star object.
        ephem (Ephem): object ephemeris. It must be an Ephemeris object.

    Returns:
        instant of CA (Time): Instant of Closest Approach
        CA (arcsec): Distance of Closest Approach
        PA (deg): Position Angle at Closest Approach
        vel (km/s): Velocity of the occultation
        dist (AU): the object geocentric distance.
    """

    delta_t = 0.05

    if type(star) != Star:
        raise ValueError('star must be a Star object')
    if type(ephem) not in [EphemKernel, EphemJPL, EphemPlanete]:
        raise ValueError('ephem must be an Ephemeris object')

    time = Time(time)
    tt = time + np.arange(-600, 600, delta_t)*u.s
    coord = star.geocentric(tt[0])
    if type(ephem) == EphemPlanete:
        ephem.fit_d2_ksi_eta(coord, log=False)

    if type(ephem) == EphemJPL:
        tt = time + np.arange(-600, 600, 4)*u.s
        ksi, eta = ephem.get_ksi_eta(tt, coord)
        dd = np.sqrt(ksi*ksi+eta*eta)
        min = np.argmin(dd)
        tt = tt[min] + np.arange(-8, 8, 0.05)*u.s

    ksi, eta = ephem.get_ksi_eta(tt, coord)
    dd = np.sqrt(ksi*ksi+eta*eta)
    min = np.argmin(dd)

    if type(ephem) == EphemPlanete:
        dist = ephem.ephem[int(len(ephem.time)/2)].distance
    else:
        dist = ephem.get_position(time).distance

    ca = np.arcsin(dd[min]*u.km/dist).to(u.arcsec)

    pa = (np.arctan2(ksi[min], eta[min])*u.rad).to(u.deg)
    if pa < 0*u.deg:
        pa = pa + 360*u.deg

    dksi = ksi[min+1]-ksi[min]
    deta = eta[min+1]-eta[min]
    vel = np.sqrt(dksi**2 + deta**2)/delta_t
    vel = vel*np.sign(dksi)*(u.km/u.s)

    return tt[min], ca, pa, vel, dist.to(u.AU)


def prediction(ephem, time_beg, time_end, mag_lim=None, step=60, divs=1, sigma=1, log=True):
    """ Predicts stellar occultations

    Parameters:
        ephem (Ephem): object ephemeris. It must be an Ephemeris object.
        time_beg (str,Time): Initial time for prediction
        time_beg (str,Time): Final time for prediction
        mag_lim (int,float): Faintest Gmag for search
        step (int, float): step, in seconds, of ephem times for search
        divs (int): number of regions the ephemeris will be splitted for better search of occultations
        sigma (int,float): ephemeris error sigma for search off-Earth.
        log (bool): To show what is being done at the moment.

    Returns:
        predict (PredictionTable): PredictionTable with the occultation params for each event
    """
    # generate ephemeris
    if type(ephem) is not EphemKernel:
        raise TypeError('At the moment prediction only works with EphemKernel')
    time_beg = Time(time_beg)
    time_end = Time(time_end)
    dt = np.arange(0, (time_end-time_beg).sec, step)*u.s
    t = time_beg + dt

    # define catalogue parameters
    kwds = {}
    kwds['columns'] = ['Source', 'RA_ICRS', 'DE_ICRS']
    kwds['row_limit'] = 10000000
    kwds['timeout'] = 600
    if mag_lim:
        kwds['column_filters'] = {"Gmag": "<{}".format(mag_lim)}
    vquery = Vizier(**kwds)

    # determine suitable divisions for star search
    radius = ephem.radius + const.R_earth

    divisions = np.array_split(np.arange(len(t)), divs)

    if log:
        print('Ephemeris was split in {} parts for better search of stars'.format(len(divisions)))

    # makes predictions for each division
    occs = []
    for i, vals in enumerate(divisions):
        nt = t[vals]
        if log:
            print('\nSearching occultations in part {}/{}'.format(i+1, len(divisions)))
            print("Generating Ephemeris between {} and {} ...".format(nt.min(), nt.max()))
        ncoord = ephem.get_position(nt)
        ra = np.mean([ncoord.ra.min().deg, ncoord.ra.max().deg])
        dec = np.mean([ncoord.dec.min().deg, ncoord.dec.max().deg])
        mindist = np.arcsin(radius/ncoord.distance).max() + sigma*np.max([ephem.error_ra.value, ephem.error_dec.value])*u.arcsec
        width = ncoord.ra.max() - ncoord.ra.min() + 2*mindist
        height = ncoord.dec.max() - ncoord.dec.min() + 2*mindist
        pos_search = SkyCoord(ra*u.deg, dec*u.deg)

        if log:
            print('Downloading stars ...')
        catalogue = vquery.query_region(pos_search, width=width, height=height, catalog='I/345/gaia2', cache=False)
        if len(catalogue) == 0:
            print('    No star found. The region is too small or VizieR is out.')
            continue
        catalogue = catalogue[0]
        if log:
            print('    {} Gaia-DR2 stars downloaded'.format(len(catalogue)))
            print('Identifying occultations ...')
        stars = SkyCoord(catalogue['RA_ICRS'], catalogue['DE_ICRS'])
        idx, d2d, d3d = stars.match_to_catalog_sky(ncoord)

        dist = np.arcsin(radius/ncoord[idx].distance) + sigma*np.max([ephem.error_ra.value, ephem.error_dec.value])*u.arcsec
        k = np.where(d2d < dist)[0]
        for ev in k:
            star = Star(code=catalogue['Source'][ev], nomad=False, log=False)
            c = star.geocentric(nt[idx][ev])
            pars = [star.code, SkyCoord(c.ra, c.dec), star.mag['G']]
            try:
                pars = np.hstack((pars, occ_params(star, ephem, nt[idx][ev])))
                occs.append(pars)
            except:
                pass

    meta = {'name': ephem.name, 'time_beg': time_beg, 'time_end': time_end, 'maglim': mag_lim, 'max_ca': mindist,
            'radius': ephem.radius.to(u.km).value, 'error_ra': ephem.error_ra.to(u.mas).value,
            'error_dec': ephem.error_dec.to(u.mas).value, 'ephem': ephem.meta['kernels']}
    if not occs:
        print('No stellar occultation was found.')
        return PredictionTable(meta=meta)
    # create astropy table with the params
    occs2 = np.transpose(occs)
    time = Time(occs2[3])
    geocentric = SkyCoord([ephem.get_position(time)])
    k = np.argsort(time)
    t = PredictionTable(
        time=time[k], coord_star=occs2[1][k], coord_obj=geocentric[k], ca=[i.value for i in occs2[4][k]],
        pa=[i.value for i in occs2[5][k]], vel=[i.value for i in occs2[6][k]], mag=occs2[2][k],
        dist=[i.value for i in occs2[7][k]], source=occs2[0][k], meta=meta)
    if log:
        print('{} occultations found.'.format(len(t)))
    return t


def xy2latlon(x, y, loncen, latcen, time):
    """ Calculates the longitude and latitude given projected positions x and y

    Parameters:
        x (int, float): Projected position in x, in the GCRS (meters)
        y (int, float): Projected position in y, in the GCRS (meters)
        loncen (int, float): Center longitude of projection (deg)
        latcen (int, float): Center latitude of projection (deg)
        time (Time): Time of refered projection

    Returns:
        lon, lat (float): Longitude and Latitude whose projection at loncen, lat results in x, y. (deg)
    """
    r = const.R_earth.to(u.m).value
    site_cen = EarthLocation(loncen*u.deg, latcen*u.deg)
    itrs_cen = site_cen.get_itrs(obstime=time)
    gcrs_cen = itrs_cen.transform_to(GCRS(obstime=time))

    z = np.array(y, ndmin=1)
    y = np.array(x, ndmin=1)
    x2 = r*r-y*y-z*z

    a = np.where(x2 >= 0.0)
    x = np.sqrt(x2[a])

    y = y[a]
    z = z[a]
    lon = np.repeat(1e+31, len(x2))
    lat = np.repeat(1e+31, len(x2))
    center_frame = SkyOffsetFrame(origin=gcrs_cen)
    if len(x) > 0:
        n = 0
        if not time.isscalar and len(time) == len(x2):
            time = time[a]
        while True:
            n += 1
            new_pos = SkyCoord(x*u.m, y*u.m, z*u.m, representation_type='cartesian', frame=center_frame[a])
            n_coord = new_pos.transform_to(GCRS(obstime=time))
            n_itrs = n_coord.transform_to(ITRS(obstime=time))
            n_site = n_itrs.earth_location
            n_site = EarthLocation(n_site.lon, n_site.lat, 0)
            itrs_site = n_site.get_itrs(obstime=time)
            gcrs_site = itrs_site.transform_to(GCRS(obstime=time))
            target1 = gcrs_site.transform_to(center_frame[a])
            if n == 4:
                lon[a] = n_site.lon.deg
                lat[a] = n_site.lat.deg
                break
            x = target1.cartesian.x.to(u.m).value
    return lon, lat


def latlon2xy(lon, lat, loncen, latcen):
    """ Calculates the projection of longitude and latitude in the loncen, latcen direction.

    Parameters:
        lon (int, float): Longitude to calculate projection
        lat (int, float): Latitude to calculate projection
        loncen (int, float): Center longitude of projection (deg)
        latcen (int, float): Center latitude of projection (deg)

    Returns:
        x, y (float): Projection of lon, lat at loncen, latcen, in the ITRS (meters)
    """
    site_cen = EarthLocation(loncen*u.deg, latcen*u.deg)
    itrs_cen = site_cen.get_itrs()

    lon = np.array(lon, ndmin=1)
    lat = np.array(lat, ndmin=1)
    site = EarthLocation(lon*u.deg, lat*u.deg, height=0*u.m)
    itrs_site = site.get_itrs()

    target = itrs_site.transform_to(SkyOffsetFrame(origin=itrs_cen))
    y = target.cartesian.y.to(u.m).value
    z = target.cartesian.z.to(u.m).value
    k = np.where(target.cartesian.x.to(u.m).value < 0.0)
    y[k] = 1e+31
    z[k] = 1e+31
    return y, z


def plot_occ_map(name, radius, coord, time, ca, pa, vel, dist, mag=0, longi=0, **kwargs):
    """ Plots map of the occultation

    Parameters:
        Required params:
        name (str): Name of the object
        radius (int, float): radius of the object (km)
        coord (str, SkyCoord): Coordinate of the star
            ("hh mm ss.sss dd mm ss.sss" or "hh.hhhhhhhh dd.dddddddd")
        time (str, Time): Instant of Closest Approach (iso or isot format)
        ca (int, float): Closest Approach Distance (arcsec)
        pa (int, float): Position Angle at C/A (deg)
        vel (int, vel): Velocity of the event (km/s)
        dist (int, float): Object distance at C/A (AU)

        Not required params (only printed in label):
        mag (int,float): Mag* = Normalized magnitude to vel=20km/s
        longi (int,float): East longitude of sub-planet point, deg, positive towards East

        Map configuration:
        nameimg (str): Change the name of the imaged saved.
        path (str): Path to a directory where to save map.
        resolution (int): Cartopy feature resolution. "1" means a resolution of "10m",
            "2" a resolution of "50m" and "3" a resolution of "100m". Default = 2
        states (bool): True to plot the state division of the countries. The states of
            some countries will only be shown depending on the resolution.
        zoom (int, float): Zooms in or out of the map.
        centermap_geo (list): Center the map given coordinates in longitude and latitude.
            It must be a list with two numbers. Default=None.
        centermap_delta (list): Displace the center of the map given displacement
            in X and Y, in km. It must be a list with two numbers. Default=None.
        centerproj (list): Rotates the Earth to show occultation with the center
            projected at a given longitude and latitude. It must be a list with two numbers
        labels (bool): Plots text above and below the map with the occultation parameters.
            Default=True.
        meridians (int): Plots lines representing the meridians for given interval. Default=30 deg
        parallels (int): Plots lines representing the parallels for given interval. Default=30 deg
        sites (dict): Plots site positions in map. It must be a python dictionary where the key is
            the name of the site, and the value is a list with longitude, latitude, delta_x,
            delta_y and color. delta_x and delta_y are displacement, in km, from the point
            of the site in the map and the name. color is the color of the point.
        site_name (bool): If True, it prints the name of the sites given, else it plots only the points
        countries (dict): Plots the names of countries. It must be a python dictionary where the key
            is the name of the country and the value is a list with longitude and latitude
            of the lower left part of the text.
        offset (list): applies an offset to the ephemeris, calculating new CA and instant of CA.
            It is a pair of delta_RA*cosDEC and delta_DEC.
        mapstyle (int): Define the color style of the map. 1 is the default black and white scale.
            2 is a colored map.
        error (int,float): Ephemeris error in mas. It plots a dashed line representing radius + error.
        ercolor (str): Changes the color of the lines of the error bar.
        ring (int,float): plots a dashed line representing the location of a ring.
            It is given in km, from the center.
        rncolor (str): Changes the color of ring lines.
        atm (int,float): plots a dashed line representing the location of an atmosphere.
            It is given in km, from the center.
        atcolor (str): Changes the color of atm lines.
        chord_delta (list): list with distances from center to plot chords
        chord_geo (2d-list): list with pairs of coordinates to plot chords
        chcolor (str): color of the line of the chords. Default: grey
        heights (list): It plots a circular dashed line showing the locations where the observer
            would observe the occultation at a given height above the horizons.
            This must be a list.
        hcolor (str): Changes the color of the height lines.
        mapsize (list): The size of figure, in cm. It must be a list with two values.
            Default = [46.0, 38.0].
        cpoints (int,float): Interval for the small points marking the center of shadow,
            in seconds. Default=60.
        ptcolor (str): Change the color of the center points.
        alpha (float): The transparency of the night shade, where 0.0 is full transparency
            and 1.0 is full black. Default = 0.2.
        fmt (str): The format to save the image. It is parsed directly by matplotlib.pyplot.
            Default = 'png'
        dpi (int): "Dots per inch". It defines the quality of the image. Default = 100.
        lncolor (str): Changes the color of the line that represents the limits of the shadow over Earth.
        outcolor (str): Changes the color of the lines that represents the limits of the shadow outside Earth
        nscale (int,float): Arbitrary scale for the size of the name of the site.
        cscale (int,float): Arbitrary scale for the name of the country.
        sscale (int,float): Arbitrary scale for the size of point of the site.
        pscale (int,float): Arbitrary scale for the size of the points that represent the center of the shadow
        arrow (bool): If true, it plots the arrow with the occultation direction.

        Comment: Only one of centermap_geo and centermap_delta can be given
    """
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature

    allowed_kwargs = ['alpha', 'arrow', 'atcolor', 'atm', 'centermap_delta', 'centermap_geo', 'centerproj',
                      'chcolor', 'chord_delta', 'chord_geo', 'countries', 'cpoints', 'cscale', 'dpi', 'ercolor',
                      'error', 'fmt', 'hcolor', 'heights', 'labels', 'lncolor', 'mapsize', 'mapstyle', 'meridians',
                      'nameimg', 'nscale', 'offset', 'outcolor', 'parallels', 'path', 'pscale', 'ptcolor',
                      'resolution', 'ring', 'rncolor', 'site_name', 'sites', 'sscale', 'states', 'zoom']
    input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)

    if not type(name) == str:
        raise TypeError('name keyword must be a string')

    radius = radius*u.km

    occs = {}
    try:
        occs['stars'] = SkyCoord(coord, frame='icrs', unit=(u.hourangle, u.degree))
    except:
        raise KeyError('"star" keyword is not in the format: "hh mm ss.sss dd mm ss.sss" or "hh.hhhhhhhh dd.dddddddd"')

    try:
        occs['datas'] = Time(time)
    except:
        raise KeyError('"time" keyword is not a iso or isot time format')
    occs['ca'] = ca*u.arcsec
    occs['posa'] = pa*u.deg
    occs['vel'] = vel*(u.km/u.s)
    occs['dist'] = dist*u.AU
    occs['magG'] = mag
    occs['longi'] = longi

    mapstyle = kwargs.get('mapstyle', 1)
    if mapstyle not in [1, 2]:
        raise ValueError('mapstyle must be 1 or 2]')

    resolution = kwargs.get('resolution', 2)
    if resolution not in [1, 2, 3]:
        raise TypeError('resolution keyword must be one of these: [1, 2, 3] where 1=10m, 2=50m and 3=100m')
    res = ['10m', '50m', '110m']
    resolution = res[resolution-1]

    nameimg = kwargs.get('nameimg', '{}_{}'.format(name, occs['datas'].isot))
    fmt = kwargs.get('fmt', 'png')
    dpi = kwargs.get('dpi', 100)
    step = kwargs.get('step', 1)
    mapsize = kwargs.get('mapsize', [46.0, 38.0])*u.cm
    erro = kwargs.get('error', None)
    ring = kwargs.get('ring', None)
    atm = kwargs.get('atm', None)
    cpoints = kwargs.get('cpoints', 60)
    states = kwargs.get('states', True)
    labels = kwargs.get('labels', True)
    meridians = kwargs.get('meridians', 30)
    parallels = kwargs.get('parallels', 30)
    nscale = kwargs.get('nscale', 1)
    cscale = kwargs.get('cscale', 1)
    sscale = kwargs.get('sscale', 1)
    pscale = kwargs.get('pscale', 1)
    heights = np.array(kwargs.get('heights'), None)
    alpha = kwargs.get('alpha', 0.2)
    centermap_geo = kwargs.get('centermap_geo', None)
    centermap_delta = kwargs.get('centermap_delta', None)
    if 'centermap_geo' in kwargs and 'centermap_delta' in kwargs:
        raise ValueError('User must give "centermap_geo" OR "centermap_delta"')
    zoom = kwargs.get('zoom', 1)
    off_ra, off_de = kwargs.get('offset', [0.0, 0.0])*u.mas
    arrow = kwargs.get('arrow', True)
    site_name = kwargs.get('site_name', True)
    path = kwargs.get('path', '.')
    chord_delta = np.array(kwargs.get('chord_delta', []), ndmin=1)*u.km
    chord_geo = kwargs.get('chord_geo', [])
    if len(chord_geo) > 0:
        try:
            b = np.array(chord_geo, ndmin=2)
            chord_geo = b.reshape(len(b), 2)
        except:
            raise ValueError('chord_geo must a set of pairs with longitude and latitude')
        chord_geo = EarthLocation(*chord_geo.T)

    sites = {}
    if 'sites' in kwargs.keys():
        if type(kwargs['sites']) == str and os.path.isfile(kwargs['sites']):
            data = np.loadtxt(kwargs['sites'], dtype={'names': ('name', 'lon', 'lat', 'offx', 'offy', 'color'),
                                                      'formats': ('S30', 'f8', 'f8', 'f8', 'f8', 'S30')},
                              delimiter=',', ndmin=1)
            for i, s in enumerate(data):
                sites[s['name'].strip().decode()] = [s['lon'], s['lat'], s['offx'], s['offy'], s['color'].strip().decode()]
        elif type(kwargs['sites']) == dict:
            sites = kwargs['sites']
        else:
            raise TypeError('sites keyword must be a file or a dictionary')

    countries = {}
    if 'countries' in kwargs.keys():
        if type(kwargs['countries']) == str and os.path.isfile(kwargs['countries']):
            data = np.loadtxt(kwargs['countries'], dtype={'names': ('name', 'lon', 'lat'), 'formats': ('S30', 'f8', 'f8')},
                              delimiter=',', ndmin=1)
            for i, c in enumerate(data):
                countries[c['name'].strip().decode()] = [c['lon'], c['lat']]
        elif type(kwargs['countries']) == dict:
            countries = kwargs['countries']
        else:
            raise TypeError('country keyword must be a file or a dictionary')

# calculates offsets
    dca = off_ra*np.sin(occs['posa']) + off_de*np.cos(occs['posa'])
    dt = (-(off_ra*np.cos(occs['posa']) - off_de*np.sin(occs['posa'])).to(u.rad)*occs['dist'].to(u.km)/np.absolute(occs['vel'])).value*u.s
    ca1 = occs['ca'] + dca
    data = occs['datas'] + dt

# define map parameters
    center_gcrs = GCRS(occs['stars'].ra, occs['stars'].dec, 1*u.R_earth, obstime=data)
    center_itrs = center_gcrs.transform_to(ITRS(obstime=data))
    center_map = center_itrs.earth_location
    centert = True
    if 'centerproj' in kwargs.keys():
        if type(kwargs['centerproj']) == EarthLocation:
            center_map = kwargs['centerproj']
        elif np.array(kwargs['centerproj']).shape == (2,):
            center_map = EarthLocation.from_geodetic(*kwargs['centerproj'], 0.0)
        else:
            raise TypeError('centerproj must be an Astropy EarthLocation Object or an array with Longitude and Latitude only')
        centert = False
    fig = plt.figure(figsize=(mapsize.to(u.imperial.inch).value))
    projection = ccrs.Orthographic(central_longitude=center_map.lon.value, central_latitude=center_map.lat.value)
    if labels:
        axf = plt.axes(projection=projection)
    else:
        axf = plt.axes([-0.001, -0.001, 1.002, 1.002], projection=projection)
    axf.set_global()

# calculates regions for zoom
    limits = None
    r = const.R_earth.to(u.m).value
    if centermap_geo is not None:
        cx, cy = latlon2xy(centermap_geo[0], centermap_geo[1], center_map.lon.value, center_map.lat.value)
        limits = [cx/1000.0, cy/1000.0]
        if np.any(np.absolute(limits) > r):
            raise ValueError('Coordinates for centermap_geo are outside the visible range.')
    elif centermap_delta is not None:
        limits = centermap_delta
    elif zoom != 1:
        limits = [0, 0]
    if limits is not None:
        dr = r/zoom
        l0 = (limits[0]*u.km).to(u.m).value
        l1 = (limits[1]*u.km).to(u.m).value
        dmsize = mapsize[0]/mapsize[1]
        if mapsize[1] < mapsize[0]:
            lx = l0 - dr*dmsize
            ux = l0 + dr*dmsize
            ly = l1 - dr
            uy = l1 + dr
        else:
            lx = l0 - dr
            ux = l0 + dr
            ly = l1 - dr/dmsize
            uy = l1 + dr/dmsize
        axf.set_xlim(lx, ux)
        axf.set_ylim(ly, uy)
        if labels and zoom > 1:
            centert = False

# plots features
    axf.coastlines(resolution=resolution, color='0.3')
    ocean = cfeature.NaturalEarthFeature('physical', 'ocean', resolution)
    land = cfeature.NaturalEarthFeature('physical', 'land', resolution)
    border = cfeature.NaturalEarthFeature('cultural', 'admin_0_countries', resolution)
    if mapstyle == 1:
        axf.add_feature(ocean, zorder=0, color='0.9')
        axf.add_feature(land, zorder=0, edgecolor='None', color='1.0')
        axf.add_feature(border, zorder=0.1, edgecolor='0.4', facecolor='None')
        axf.add_feature(cfeature.RIVERS, zorder=0, edgecolor='0.7')
        axf.add_feature(cfeature.LAKES, zorder=0, color='0.7')
        ptcolor = 'black'
        lncolor = 'blue'
        ercolor = 'blue'
        rncolor = 'blue'
        atcolor = 'blue'
        outcolor = 'red'
        hcolor = 'black'
        chcolor = 'gray'
    elif mapstyle == 2:
        axf.add_feature(ocean, zorder=0, facecolor=cfeature.COLORS['water'])
        axf.add_feature(land, zorder=0, edgecolor='None', facecolor=cfeature.COLORS['land'])
        axf.add_feature(border, zorder=0, edgecolor='0.5', facecolor=cfeature.COLORS['land'])
        axf.add_feature(border, zorder=0.1, edgecolor='0.5', facecolor='None')
        axf.add_feature(cfeature.RIVERS, zorder=0)
        axf.add_feature(cfeature.LAKES, zorder=0)
        ptcolor = 'red'
        lncolor = 'blue'
        ercolor = 'red'
        rncolor = 'black'
        atcolor = 'black'
        outcolor = 'red'
        hcolor = 'black'
        chcolor = 'gray'
    if states:
        states_r = cfeature.NaturalEarthFeature('cultural', 'admin_1_states_provinces', resolution)
        axf.add_feature(states_r, zorder=0, edgecolor='0.6', facecolor='None')

    gl = axf.gridlines(xlocs=np.arange(-180, 180.001, meridians), ylocs=np.arange(-90, 90.001, parallels))
    gl.n_steps = 180
    sun = get_sun(data)
    sun_lat = sun.dec
    sun_lon = sun.ra - data.sidereal_time('mean', 'greenwich')
    pole_lon = sun_lon.deg
    pole_lat = sun_lat.deg
    proj_sun = ccrs.Orthographic(central_longitude=pole_lon+180, central_latitude=-pole_lat)
    bordx = r*np.cos(np.arange(0, 361, 0.5)*u.deg)
    bordy = r*np.sin(np.arange(0, 361, 0.5)*u.deg)
    axf.fill(bordx, bordy, transform=proj_sun, linewidth=0, color='black', alpha=alpha)
    axf.fill(bordx*np.cos(18*u.deg), bordy*np.cos(18*u.deg), transform=proj_sun, linewidth=0, color='black', alpha=alpha)

    ptcolor = kwargs.get('ptcolor', ptcolor)
    lncolor = kwargs.get('lncolor', lncolor)
    ercolor = kwargs.get('ercolor', ercolor)
    rncolor = kwargs.get('rncolor', rncolor)
    atcolor = kwargs.get('atcolor', atcolor)
    outcolor = kwargs.get('outcolor', outcolor)
    hcolor = kwargs.get('hcolor', hcolor)
    chcolor = kwargs.get('chcolor', chcolor)

# calculates path
    vec = np.arange(0, int(8000/(np.absolute(occs['vel'].value))), step)
    vec = np.sort(np.concatenate((vec, -vec[1:]), axis=0))
    pa = Angle(occs['posa'])
    pa.wrap_at('180d', inplace=True)
    if pa > 90*u.deg:
        paplus = pa - 180*u.deg
    elif pa < -90*u.deg:
        paplus = pa + 180*u.deg
    else:
        paplus = pa
    deltatime = vec*u.s
    datas1 = data + deltatime
    centers_gcrs = GCRS(np.repeat(occs['stars'].ra, len(datas1)), np.repeat(occs['stars'].dec, len(datas1)),
                        1*u.R_earth, obstime=datas1)
    centers_itrs = centers_gcrs.transform_to(ITRS(obstime=datas1))
    centers = centers_itrs.earth_location

    dista = (occs['dist'].to(u.km)*ca1.to(u.rad)).value*u.km
    ax = dista*np.sin(pa) + (deltatime*occs['vel'])*np.cos(paplus)
    by = dista*np.cos(pa) - (deltatime*occs['vel'])*np.sin(paplus)

    ax2 = ax - (radius)*np.sin(paplus)
    by2 = by - (radius)*np.cos(paplus)
    lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
    j = np.where(lon1 < 1e+30)
    axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=lncolor)
    j = np.where(lon1 > 1e+30)
    if 'centerproj' not in kwargs:
        plt.plot(ax2[j].to(u.m).value, by2[j].to(u.m).value, color=outcolor, clip_on=(not centert), zorder=-0.2)

    ax3 = ax + (radius)*np.sin(paplus)
    by3 = by + (radius)*np.cos(paplus)
    lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
    j = np.where(lon2 < 1e+30)
    axf.plot(lon2[j], lat2[j], transform=ccrs.Geodetic(),  color=lncolor)
    j = np.where(lon2 > 1e+30)
    if 'centerproj' not in kwargs:
        plt.plot(ax3[j].to(u.m).value, by3[j].to(u.m).value, color=outcolor, clip_on=(not centert), zorder=-0.2)

# plots chords_delta
    for val in chord_delta:
        ax2 = ax + val*np.sin(paplus)
        by2 = by + val*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=chcolor)

# plots chords_geo
    for coord_geo in chord_geo:
        xt, yt = latlon2xy(coord_geo.lon.deg, coord_geo.lat.deg, centers.lon.value, centers.lat.value)*u.m
        val = np.sqrt((xt-ax)**2 + (yt-by)**2)
        k = val.argmin()
        ang = np.arctan2((yt-by)[k], (xt-ax)[k])
        val = np.sign(np.sin(ang))*val[k]
        ax2 = ax + val*np.sin(paplus)
        by2 = by + val*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], transform=ccrs.Geodetic(), color=chcolor)

# plots error
    if erro is not None:
        err = erro*u.mas
        errd = (occs['dist'].to(u.km)*err.to(u.rad)).value*u.km
        ax2 = ax - errd*np.sin(paplus) - radius*np.sin(paplus)
        by2 = by - errd*np.cos(paplus) - radius*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(),  color=ercolor)

        ax3 = ax + errd*np.sin(paplus) + radius*np.sin(paplus)
        by3 = by + errd*np.cos(paplus) + radius*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(),  color=ercolor)

# plots ring
    if ring is not None:
        rng = ring*u.km
        ax2 = ax - rng*np.sin(paplus)
        by2 = by - rng*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(),  color=rncolor)

        ax3 = ax + rng*np.sin(paplus)
        by3 = by + rng*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(),  color=rncolor)

# plots atm
    if atm is not None:
        atmo = atm*u.km
        ax2 = ax - atmo*np.sin(paplus)
        by2 = by - atmo*np.cos(paplus)
        lon1, lat1 = xy2latlon(ax2.to(u.m).value, by2.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon1 < 1e+30)
        axf.plot(lon1[j], lat1[j], '--', transform=ccrs.Geodetic(),  color=atcolor)

        ax3 = ax + atmo*np.sin(paplus)
        by3 = by + atmo*np.cos(paplus)
        lon2, lat2 = xy2latlon(ax3.to(u.m).value, by3.to(u.m).value, centers.lon.value, centers.lat.value, datas1)
        j = np.where(lon2 < 1e+30)
        axf.plot(lon2[j], lat2[j], '--', transform=ccrs.Geodetic(),  color=atcolor)

# plots center points
    vec = np.arange(0, int(8000/(np.absolute(occs['vel'].value))), cpoints)
    deltatime = np.sort(np.concatenate((vec, -vec[1:]), axis=0))*u.s
    axc = dista*np.sin(pa) + (deltatime*occs['vel'])*np.cos(paplus)
    byc = dista*np.cos(pa) - (deltatime*occs['vel'])*np.sin(paplus)
    plt.plot(axc.to(u.m).value, byc.to(u.m).value, 'o', color=ptcolor, clip_on=(not centert),
             markersize=mapsize[0].value*pscale*8.0/46.0, zorder=-0.2)

    datas2 = data + deltatime
    centers_p_gcrs = GCRS(np.repeat(occs['stars'].ra, len(datas2)), np.repeat(occs['stars'].dec, len(datas2)),
                          1*u.R_earth, obstime=datas2)
    centers_p_itrs = centers_p_gcrs.transform_to(ITRS(obstime=datas2))
    centers_p = centers_p_itrs.earth_location
    clon1, clat1 = xy2latlon(axc.to(u.m).value, byc.to(u.m).value, centers_p.lon.value, centers_p.lat.value, datas2)
    j = np.where(clon1 < 1e+30)
    axf.plot(clon1[j], clat1[j], 'o', transform=ccrs.Geodetic(), color=ptcolor, clip_on=True,
             markersize=mapsize[0].value*pscale*8.0/46.0)

    datas1 = data + deltatime
    center_gcrs = GCRS(np.repeat(occs['stars'].ra, 1), np.repeat(occs['stars'].dec, 1),
                       1*u.R_earth, obstime=data)
    center_itrs = center_gcrs.transform_to(ITRS(obstime=data))
    center = center_itrs.earth_location
    xp = [(dista.to(u.m)*np.sin(pa)).value]
    yp = [(dista.to(u.m)*np.cos(pa)).value]
    loncen, latcen = xy2latlon(xp, yp, center.lon.value, center.lat.value, data)
    j = np.where(loncen < 1e+30)
    if len(j) > 0:
        axf.plot(loncen[j], latcen[j], 'o', transform=ccrs.Geodetic(), color=ptcolor, clip_on=True,
                 markersize=mapsize[0].value*pscale*24.0/46.0)
    elif not centert:
        plt.plot(xp, yp, 'o', color=ptcolor, clip_on=False, markersize=mapsize[0].value*pscale*24.0/46.0)

# plots the heights
    if 'heights' in kwargs.keys():
        for h in heights:
            plt.plot(bordx*np.cos(h*u.deg), bordy*np.cos(h*u.deg), linestyle='dotted', color=hcolor)

# plots the the direction arrow
    if arrow:
        if limits is None:
            plt.quiver(5500000, -5500000, (np.sin(paplus+90*u.deg)*np.sign(occs['vel'])).value,
                       (np.cos(paplus+90*u.deg)*np.sign(occs['vel'])).value, width=0.005)
        else:
            plt.quiver(lx + (ux-lx)*0.9, ly + (uy-ly)*0.1, (np.sin(paplus+90*u.deg)*np.sign(occs['vel'])).value,
                       (np.cos(paplus+90*u.deg)*np.sign(occs['vel'])).value, width=0.005, zorder=1.3)

# plots the countries names
    for country in countries.keys():
        plt.text(countries[country][0], countries[country][1], country, transform=ccrs.Geodetic(),
                 weight='bold', color='grey', fontsize=30*cscale, family='monospace')

# plots the sites
    for site in sites.keys():
        s = EarthLocation.from_geodetic(sites[site][0], sites[site][1], 0.0*u.km)
        axf.plot(s.lon.deg, s.lat.deg, 'o', transform=ccrs.Geodetic(),
                 markersize=mapsize[0].value*sscale*10.0/46.0, color=sites[site][4])
        if site_name:
            xt, yt = latlon2xy(s.lon.deg, s.lat.deg, center_map.lon.value, center_map.lat.value)
            axf.text(xt + sites[site][2]*1000, yt+sites[site][3]*1000, site, weight='bold',
                     fontsize=25*nscale, family='monospace')

# Define the title and label of the output
    title = ('Object        Diam   Tmax   dots <> ra_offset_dec\n'
             '{:10s} {:4.0f} km  {:5.1f}s  {:02d} s <>{:+6.1f} {:+6.1f} \n'.
             format(name, 2*radius.value, (2*radius/np.absolute(occs['vel'])).value,
                    cpoints, off_ra.value, off_de.value))
    labelx = ("\n year-m-d    h:m:s UT     ra__dec__J2000__candidate    C/A    P/A    vel   Delta   G*  long\n"
              "{}  {:02d} {:02d} {:07.4f} {:+03d} {:02d} {:06.3f} {:6.3f} {:6.2f} {:6.2f}  {:5.2f} {:5.1f}  {:3.0f}".
              format(data.iso, int(occs['stars'].ra.hms.h), int(occs['stars'].ra.hms.m), occs['stars'].ra.hms.s,
                     int(occs['stars'].dec.dms.d), np.absolute(int(occs['stars'].dec.dms.m)),
                     np.absolute(occs['stars'].dec.dms.s), ca1.value, occs['posa'].value,
                     occs['vel'].value, occs['dist'].value, occs['magG'], occs['longi']))

# plots the map
    if labels:
        axf.set_title(title, family='monospace', weight='bold', fontsize=22)
        axf.text(0.5, -0.1, labelx, va='bottom', ha='center', rotation='horizontal', rotation_mode='anchor',
                 transform=axf.transAxes, family='monospace', weight='bold', fontsize=22)
    filepath = os.path.join(path, '{}.{}'.format(nameimg, fmt))
    plt.savefig(filepath, format=fmt, dpi=dpi)
    print('{}.{} generated'.format(nameimg, fmt))
    plt.clf()
    plt.close()



import numpy as np
import matplotlib.pylab as pl
import astropy.units as u
from astropy.timeseries import BoxLeastSquares
from astropy.time import Time
import scipy.special as scsp
from scipy.odr import odrpack as odr
from scipy.odr import models
from .extra import ChiSquare
from sora.config import input_tests
import os
import warnings
from sora.config.decorators import deprecated_alias


warnings.simplefilter('always', UserWarning)

@deprecated_alias(lambida='bandpass')  # remove this line for v1.0
def calc_fresnel(distance, bandpass):
    """ Calculates the Fresnel scale.
        (Fresnel Scale = square root of half the multiplication of wavelength and object distance.)

    Parameters:
        distance  (int, float, array): distances, in km.
        bandpass  (int, float, array): wavelength, in km.

    Returns:
        fresnel_scale  (float, array): Fresnel scale, in km
    """
    return np.sqrt(bandpass*distance/2)


def fit_pol(x, y, deg):
    """ Fits a n-degree polynom to the data.

    Parameters:
        x (array): x-values
        y (array): y-values
        deg (int): degree of the polynom

    Returns:
        param (array): Array with the fitted values
        param_err (array): Array with the errors of the fitted values
    """
    func = models.polynomial(deg)
    mydata = odr.Data(x, y)
    myodr = odr.ODR(mydata, func, maxit=200)
    myodr.set_job(fit_type=2)
    fit = myodr.run()
    param = fit.beta[::-1]
    param_err = fit.sd_beta[::-1]
    return param, param_err


class LightCurve():
    __names = []

    @deprecated_alias(lambda_0='central_bandpass', delta_lambda='delta_bandpass')  # remove this line for v1.0
    def __init__(self, name, **kwargs):
        """ Defines a Light Curve

        Parameters:
            name (str): The name of the LightCurve. (required)
                Each time an LightCurve object is defined the name must be different.
            tref (Time,str,float): Instant of reference.
                Format: Julian Date, string in ISO format or Time object.
                Required only if LightCurve have input fluxes and given time is not in Julian Date.
            central_bandpass (int,float): The center band pass of the detector used in observation.
                Value in microns (not required). Default=0.7
            delta_bandpass (int,float): The band pass width of the detector used in observation.
                Value in microns (not required). Default=0.3
            exptime (int,float): The exposure time of the observation.
                NOT required in cases 2, 3 and 4 below
                Required in case 1 below

            Input data must be one of the 4 options below:

            1) Input file with time and flux
                file (str): a file with the time and flux.
                    A third column with the error in flux can also be given.
                usecols (int, tuple, array): Which columns to read, with the 
                    first being the time, the seconds the flux and third the flux error [optional].

            2) IF file is not given:
                time: time must be a list of times, in seconds from tref,
                    or Julian Date, or a Time object.
                flux: flux must be a list of fluxes. It must have the
                    same lenght as time.
                dflux: if file not given, dflux must be a list of fluxes errors.
                    It must have the same lenght as time. (not required)

            IF time and flux are not given.
            3) For a positive occultation
                immersion: The instant of immersion.
                emersion: The instant of emersion
                immersion_err: Immersion time uncertainty
                emersion_err: Emersion time uncertainty

            4) For a negative occultation
                initial_time: The initial time of observation
                end_time: The end time of observation.

        Examples: The user can provide one of the followings:

            LightCurve(name, flux, time, exptime) # dflux can also be given
            LightCurve(name, file, exptime) # dflux can also be given
            LightCurve(name, immersion, immersion_err, emersion, emersion_err)
            LightCurve(name, initial_time, end_time)
        """
        allowed_kwargs = ['emersion', 'emersion_err', 'immersion', 'immersion_err', 'initial_time', 'end_time',
                          'file', 'time', 'flux', 'exptime', 'central_bandpass', 'delta_bandpass', 'tref', 'dflux', 'usecols']
        input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)
        input_done = False
        self.dflux = None
        self.__name = name
        self.flux = None
        self.time_model = None
        if self.__name in self.__names:
            raise ValueError('name {} already defined for another LightCurve object. Please choose a different one.'.
                             format(self.__name))
        if 'tref' in kwargs:
            self.tref = kwargs['tref']
        if 'immersion' in kwargs:
            self.immersion = kwargs['immersion']
            self.immersion_err = kwargs.get('immersion_err', 0.0)
            input_done = True
        if 'emersion' in kwargs:
            self.emersion = kwargs['emersion']
            self.emersion_err = kwargs.get('emersion_err', 0.0)
            input_done = True
        if 'initial_time' in kwargs and 'end_time' in kwargs:
            self.initial_time = kwargs['initial_time']
            self.end_time = kwargs['end_time']
            input_done = True
        if not input_done:
            try:
                self.set_flux(**kwargs)
            except:
                raise ValueError('No allowed input conditions satisfied. Please refer to the tutorial.')
        self.lambda_0 = kwargs.get('central_bandpass', 0.70)
        self.delta_lambda = kwargs.get('delta_bandpass', 0.30)
        self.dt = 0.0
        self.__names.append(self.__name)

    @property
    def fresnel_scale(self):
        lamb = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist = self.dist*u.au.to('km')
        fresnel_scale_1 = calc_fresnel(dist, lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist, lamb+dlamb/2.0)
        fresnel_scale = (fresnel_scale_1 + fresnel_scale_2)/2.0
        return fresnel_scale

    @property
    def central_bandpass(self):
        return self.lambda_0

    @property
    def delta_bandpass(self):
        return self.delta_lambda

    @property
    def name(self):
        return self.__name

    @property
    def tref(self):
        if hasattr(self, '_tref'):
            return self._tref
        else:
            raise AttributeError("'LightCurve' object has no attribute 'tref'")

    @tref.setter
    def tref(self, value):
        if type(value) in [int, float]:
            self.tref = Time(value, format='jd')
        else:
            try:
                self._tref = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by tref'.format(value))

    @property
    def immersion(self):
        if hasattr(self, '_immersion'):
            return self._immersion + self.dt*u.s
        else:
            raise AttributeError('The immersion time was not fitted or instanciated.')

    @immersion.setter
    def immersion(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.immersion = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.immersion = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._immersion = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by immersion'.format(value))

    @property
    def emersion(self):
        if hasattr(self, '_emersion'):
            return self._emersion + self.dt*u.s
        else:
            raise AttributeError('The emersion time was not fitted or instanciated.')

    @emersion.setter
    def emersion(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.emersion = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.emersion = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._emersion = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by emersion'.format(value))

    @property
    def initial_time(self):
        if hasattr(self, '_initial_time'):
            return self._initial_time
        else:
            raise AttributeError("'LightCurve' object has no attribute 'initial_time'")

    @initial_time.setter
    def initial_time(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.initial_time = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.initial_time = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._initial_time = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by initial_time'.format(value))

    @property
    def end_time(self):
        if hasattr(self, '_end_time'):
            return self._end_time
        else:
            raise AttributeError("'LightCurve' object has no attribute 'end_time'")

    @end_time.setter
    def end_time(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.end_time = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.end_time = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._end_time = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by end_time'.format(value))

    @property
    def time_mean(self):
        if hasattr(self, '_immersion') and hasattr(self, '_emersion'):
            return Time((self.immersion.jd + self.emersion.jd)/2, format='jd')
        else:
            return Time((self.initial_time.jd + self.end_time.jd)/2, format='jd')

    @property
    def time(self):
        try:
            return (self._time - self.tref).sec
        except:
            raise AttributeError("'LightCurve' object has no attribute 'time'")

    def check_names(self):
        return self.__names

    def set_flux(self, **kwargs):
        """ Sets the flux for the LightCurve

        Parameters:
            exptime (int,float): The exposure time of the observation. (required)
            file (str): a file with the time and flux in the first and second columns, respectively.
                A third column with error in flux can also be given.
            time: if file not given, time must be a list of times, in seconds from tref, or Julian Date,
                or a Time object.
            flux: if file not given, flux must be a list of fluxes. It must have the same lenght as time.
            dflux: if file not given, dflux must be a list of fluxes errors.
                It must have the same lenght as time.
            tref (Time,str,float): Instant of reference. It can be in Julian Date, string in ISO format
                or Time object.
            usecols (int, tuple, array): Which columns to read, with the 
                    first being the time, the seconds the flux and third the flux error [optional].
        """
        input_done = False
        usecols = None
        if 'usecols' in kwargs:
            usecols = kwargs['usecols']
        if 'file' in kwargs:
            if not os.path.isfile(kwargs['file']):
                raise ValueError('{} not found'.format(kwargs['file']))
            if usecols is not None:
                if len(usecols) == 2:
                    time, self.flux = np.loadtxt(kwargs['file'], usecols=usecols, unpack=True)
                elif len(usecols) == 3:
                    time, self.flux, self.dflux = np.loadtxt(kwargs['file'], usecols=usecols, unpack=True)
                else:
                    raise ValueError('usecols should have 2 or 3 values')
            else:
                try:
                    time, self.flux, self.dflux = np.loadtxt(kwargs['file'], usecols=[0, 1, 2], unpack=True)
                except:
                    pass
                try:
                    time, self.flux = np.loadtxt(kwargs['file'], usecols=[0, 1], unpack=True)
                except:
                    pass
            if hasattr(self, 'flux'):
                self.flux_obs = self.flux
            if not hasattr(self, 'flux_obs'):
                raise ValueError('Input file must have 2 or 3 columns')
            input_done = True
        if 'time' in kwargs and 'flux' in kwargs:
            if input_done:
                raise ValueError('Only one type of input can be given. Please refer to the tutorial.')
            self.flux = kwargs['flux']
            time = kwargs['time']
            if len(self.flux) != len(time):
                raise ValueError('time and flux must have the same length')
            if 'dflux' in kwargs:
                self.dflux = kwargs['dflux']
                if len(self.flux) != len(self.dflux):
                    raise ValueError('dflux must have the same length as flux and time')
            input_done = True
        if 'exptime' not in kwargs:
            raise ValueError('exptime not defined')
        if kwargs['exptime'] <= 0:
            raise ValueError('Exposure time can not be zero or negative')
        else:
            self.exptime = kwargs['exptime']
        if 'tref' in kwargs:
            self.tref = kwargs['tref']
        if 'time' in locals():
            if type(time) == Time:
                if not hasattr(self, 'tref'):
                    self.tref = Time(time[0].iso.split(' ')[0] + ' 00:00:00.000')
            elif all(time > 2400000):
                time = Time(time, format='jd')
                if not hasattr(self, 'tref'):
                    self.tref = Time(time[0].iso.split(' ')[0] + ' 00:00:00.000')
            elif not hasattr(self, 'tref'):
                raise ValueError('tref must be given')
            else:
                time = self.tref + time*u.s
            order = np.argsort(time)
            self._time = time[order]
            self.model = np.ones(len(time))
            self.flux = self.flux[order]
            self.flux_obs = self.flux
            if self.dflux is not None:
                self.dflux = self.dflux[order]
            self.initial_time = np.min(time)
            self.end_time = np.max(time)
            self.cycle = np.median(time[1:] - time[:-1]).sec
            if self.cycle < self.exptime:
                warnings.warn('Exposure time ({:0.4f} seconds) higher than Cycle time ({:0.4f} seconds)'.
                              format(self.exptime, self.cycle))

    def set_vel(self, vel):
        """ Sets the occultation velocity

        Parameters:
            vel (int,float): velocity in km/s
        """
        if type(vel) == u.quantity.Quantity:
            vel = vel.to(u.km/u.s).value
        elif type(vel) in [float, int]:
            pass
        else:
            raise TypeError('vel must be an integer, a float or an Astropy Unit object')
        self.vel = np.absolute(vel)

    def set_dist(self, dist):
        """ Sets the object distance

        Parameters:
            dist (int,float): object distance in km
        """
        if type(dist) == u.quantity.Quantity:
            dist = dist.to(u.AU).value
        elif type(dist) in [float, int]:
            pass
        else:
            raise TypeError('dist must be an integer, a float or an Astropy Unit object')
        self.dist = dist

    def set_star_diam(self, d_star):
        """ Sets the star diameter

        Parameters:
            d_star (float): star diameter, in km
        """
        if type(d_star) == u.quantity.Quantity:
            d_star = d_star.to(u.km).value
        elif type(d_star) in [float, int]:
            pass
        else:
            raise TypeError('d_star must be an integer, a float or an Astropy Unit object')
        self.d_star = d_star

    @deprecated_alias(lambda_0='central_bandpass', delta_lambda='delta_bandpass')  # remove this line for v1.0
    def set_filter(self, central_bandpass, delta_bandpass):
        """ Sets the filter bandwidth in microns

        Parameters:
            central_bandpass (float): center band in microns
            delta_bandpass (float): bandwidth in microns
        """
        if type(central_bandpass) == u.quantity.Quantity:
            central_bandpass = central_bandpass.to(u.micrometer).value
        elif type(central_bandpass) in [float]:
            pass
        else:
            raise TypeError('central_bandpass must be a float or an Astropy Unit object')
        self.lambda_0 = central_bandpass
        if type(delta_bandpass) == u.quantity.Quantity:
            delta_bandpass = delta_bandpass.to(u.micrometer).value
        elif type(delta_bandpass) in [float]:
            pass
        else:
            raise TypeError('delta_bandpass must be a float or an Astropy Unit object')
        self.delta_lambda = delta_bandpass

    def calc_magnitude_drop(self, mag_star, mag_obj):
        """ Determines the magnitude drop of the occultation

        Parameters:
            mag_star (int,float): Star magnitude.
            mag_obj (int,float): Object apparent magnitude to the date.

        Returns:
            mag_drop (float): Magnitude drop for the given magnitudes
            bottom_flux (float): Normalized bottom flux for the given magnitudes
        """
        contrast = 1/(1+(10**((mag_star-mag_obj)*0.4)))
        mag_combined = mag_star-(2.5*(np.log10(1/contrast)))
        mag_drop = mag_obj - mag_combined
        bottom_flux = 10**((mag_combined - mag_obj)*0.4)
        self.mag_drop = mag_drop
        self.bottom_flux = bottom_flux
        return

    def normalize(self, poly_deg=None, mask=None, flux_min=0.0, flux_max=1.0, plot=False):
        """ Returns the fresnel scale.

        Parameters:
            poly_deg (int): degree of the polynom to be fitted
            mask (array of Bolleans): which values to be fitted
            flux_min (int,float): event flux to be setted as 0.0
            flux_max (int,float): baseline flux to be setted as 1.0
            plot (Bollean): If True plot the steps for visual aid
        """
        # Create a mask where the polynomial fit will be done
        if not all(self.flux):
            raise ValueError('Normalization is only possible when a LightCurve is instatiated with time and flux.')
        self.reset_flux()
        lc_flux = (self.flux - flux_min)/(flux_max-flux_min)
        if mask is None:
            preliminar_occ = self.occ_detect(maximum_duration=((self.end_time - self.initial_time).value*u.d.to('s'))/3)
            tmax = preliminar_occ['emersion_time']+1.00*preliminar_occ['occultation_duration']
            tmin = preliminar_occ['immersion_time']-1.00*preliminar_occ['occultation_duration']
            chord = preliminar_occ['occultation_duration']
            mask = np.invert((self.time > tmin-(chord/2)) & (self.time < tmax+(chord/2)))
        norm_time = (self.time - self.time.min())/(self.time.max()-self.time.min())
        if poly_deg is not None:
            n = poly_deg
            p, err = fit_pol(norm_time[mask], lc_flux[mask], n)
            flux_poly_model = np.zeros(len(norm_time))
            for ii in np.arange(n+1):
                flux_poly_model = flux_poly_model + p[ii]*(norm_time**(n-ii))
            if plot:
                pl.plot(norm_time[mask], lc_flux[mask], 'k.-')
                pl.plot(norm_time[mask], flux_poly_model[mask], 'r-')
                pl.title('Polynomial degree = {}'.format(n), fontsize=15)
                pl.show()
        if poly_deg is None:
            n = 0
            p, err = fit_pol(norm_time[mask], lc_flux[mask], n)
            flux_poly_model = np.zeros(len(norm_time))
            for ii in np.arange(n+1):
                flux_poly_model += p[ii]*(norm_time**(n-ii))
            if plot:
                pl.plot(norm_time[mask], lc_flux[mask], 'k.-')
                pl.plot(norm_time[mask], flux_poly_model[mask], 'r-')
                pl.title('Polynomial degree = {}'.format(n), fontsize=15)
                pl.show()
            for nn in np.arange(1, 10):
                p, err = fit_pol(norm_time[mask], lc_flux[mask], nn)
                flux_poly_model_new = np.zeros(len(norm_time))
                for ii in np.arange(nn+1):
                    flux_poly_model_new += p[ii]*(norm_time**(nn-ii))
                F = np.var(flux_poly_model[mask]-lc_flux[mask])/np.var(flux_poly_model_new[mask]-lc_flux[mask])
                if F > 1.05:
                    flux_poly_model = flux_poly_model_new.copy()
                    n = nn
                    if plot:
                        pl.plot(norm_time[mask], lc_flux[mask], 'k.-')
                        pl.plot(norm_time[mask], flux_poly_model[mask], 'r-')
                        pl.title('Polynomial degree = {}'.format(nn), fontsize=15)
                        pl.show()
                else:
                    print('Normalization using a {} degree polynom'.format(n))
                    print('There is no improvement with a {} degree polynom'.format(n+1))
                    break
        self.flux = lc_flux/flux_poly_model
        self.normalizer_flux = flux_poly_model
        self.normalizer_mask = mask
        return

    def reset_flux(self):
        """ Resets flux for original values
        """
        try:
            self.flux = self.flux_obs
        except:
            raise ValueError('Reset is only possible when a LightCurve is instatiated with time and flux.')
        return

    def occ_model(self, immersion_time, emersion_time, opacity, mask, npt_star=12,
                  time_resolution_factor=10, flux_min=0, flux_max=1):
        """ Returns the modelled light curve considering fresnel difraction, star diameter and intrumental response.

        Parameters:
            immersion_time (int, float): Immersion time, in seconds.
            emersion_time (int, float): Emersion time, in seconds.
            opacity (int, float): Opacity. Opaque = 1.0, transparent = 0.0
            mask (array with Booleans): Mask with True values to be computed
            npt_star  (int): Number of subdivisions for computing the star size's effects. Default=12
            time_resolution_factor (int,float): Steps for fresnel scale used for modelling the light curve.
                Default=10*fresnel scale.
            flux_min (int,float): Bottom flux (only object). Default=0.0
            flux_max (int,float): Base flux (object plus star). Default=1.0
        """
        # Computing the fresnel scale
        lamb = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist = self.dist*u.au.to('km')
        vel = np.absolute(self.vel)
        time_obs = self.time[mask]
        fresnel_scale_1 = calc_fresnel(dist, lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist, lamb+dlamb/2.0)
        fresnel_scale = (fresnel_scale_1 + fresnel_scale_2)/2.0
        time_resolution = (np.min([fresnel_scale/vel, self.exptime]))/time_resolution_factor

        # Creating a high resolution curve to compute fresnel difraction, stellar diameter and instrumental integration
        time_model = np.arange(time_obs.min()-5*self.exptime, time_obs.max()+5*self.exptime, time_resolution)

        # Changing X: time (s) to distances in the sky plane (km), considering the tangential velocity (vel in km/s)
        x = time_model*vel
        x01 = immersion_time*vel
        x02 = emersion_time*vel

        # Computing fresnel diffraction for the case where the star size is negligenciable
        flux_fresnel_1 = self.__bar_fresnel(x, x01, x02, fresnel_scale_1, opacity)
        flux_fresnel_2 = self.__bar_fresnel(x, x01, x02, fresnel_scale_2, opacity)
        flux_fresnel = (flux_fresnel_1 + flux_fresnel_2)/2.
        flux_star = flux_fresnel.copy()
        if (self.d_star > 0):
            # Computing fresnel diffraction for the case where the star size is not negligenciable
            resolucao = self.d_star/npt_star
            flux_star_1 = np.zeros(len(time_model))
            flux_star_2 = np.zeros(len(time_model))
            # Computing stellar diameter only near the immersion or emersion times
            star_diam = (np.absolute(x - x01) < 3*self.d_star) + (np.absolute(x - x02) < 3*self.d_star)
            p = np.arange(-npt_star, npt_star)*resolucao
            coeff = np.sqrt(np.absolute(self.d_star**2 - p**2))
            for ii in np.where(star_diam == True)[0]:
                xx = x[ii] + p
                flux1 = self.__bar_fresnel(xx, x01, x02, fresnel_scale_1, opacity)
                flux2 = self.__bar_fresnel(xx, x01, x02, fresnel_scale_2, opacity)
                flux_star_1[ii] = np.sum(coeff*flux1)/coeff.sum()
                flux_star_2[ii] = np.sum(coeff*flux2)/coeff.sum()
                flux_star[ii] = (flux_star_1[ii] + flux_star_2[ii])/2.
        flux_inst = np.zeros(len(time_obs))
        for i in range(len(time_obs)):
            event_model = (time_model > time_obs[i]-self.exptime/2.) & (time_model < time_obs[i]+self.exptime/2.)
            flux_inst[i] = (flux_star[event_model]).mean()
        self.model[mask] = flux_inst*(flux_max - flux_min) + flux_min
        self.time_model = time_model
        self.model_star = flux_star*(flux_max - flux_min) + flux_min
        self.model_fresnel = flux_fresnel*(flux_max - flux_min) + flux_min
        ev_model = (time_model > immersion_time) & (time_model < emersion_time)
        flux_box = np.ones(len(time_model))
        flux_box[ev_model] = (1-opacity)**2
        flux_box = flux_box*(flux_max - flux_min) + flux_min
        self.model_geometric = flux_box
        self.baseflux = flux_max
        self.bottomflux = flux_min
        return

    def occ_lcfit(self, **kwargs):
        """ Monte Carlo chi square fit for occultations lightcurve.

        Parameters:
            tmin (int,float): Minimum time to consider in the fit procedure, in seconds
            tmax (int,float): Maximum time to consider in the fit procedure, in seconds
            flux_min (int,float): Bottom flux (only object). Default=0.0
            flux_max (int,float): Base flux (object plus star). Default=1.0
            immersion_time (int, float): Initial guess for immersion time, in seconds.
            emersion_time (int, float): Initial guess for emersion time, in seconds.
            opacity (int, float): Initial guess for opacity. Opaque=1.0, transparent=0.0. Default=1.0
            delta_t (int, float): Interval to fit immersion or emersion time
            dopacity (int, float): Interval to fit opacity. Default=0
            loop (int): Number of tests to be done. Default=10000

        Returns:
            chi2 (ChiSquare): ChiSquare object
        """
        allowed_kwargs = ['tmin', 'tmax', 'flux_min', 'flux_max', 'immersion_time', 'emersion_time', 'opacity',
                          'delta_t', 'dopacity', 'loop']
        input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)

        if not hasattr(self, 'flux'):
            raise ValueError('Fit curve is only possible when a LightCurve is instatiated with time and flux.')
        delta_t = 2*self.cycle
        loop = kwargs.get('loop', 10000)
        t_i = np.zeros(loop)
        t_e = np.zeros(loop)
        tmax = self.time.max()
        tmin = self.time.min()
        immersion_time = tmin - self.exptime
        do_immersion = False
        emersion_time = tmax + self.exptime
        do_emersion = False
        opacity = kwargs.get('opacity', 1.0)
        delta_opacity = 0.0
        do_opacity = False
        if ('immersion_time' not in kwargs) and ('emersion_time' not in kwargs):
            preliminar_occ = self.occ_detect()
            immersion_time = preliminar_occ['immersion_time']
            do_immersion = True
            emersion_time = preliminar_occ['emersion_time']
            do_emersion = True
            delta_t = 5*preliminar_occ['time_err']
            tmax = emersion_time+2*preliminar_occ['occultation_duration']
            tmin = immersion_time-2*preliminar_occ['occultation_duration']
            if 2*preliminar_occ['occultation_duration'] < 10*self.cycle:
                tmax = emersion_time + 10*self.cycle
                tmin = immersion_time - 10*self.cycle
        if 'tmax' in kwargs:
            tmax = kwargs['tmax']
        if 'tmin' in kwargs:
            tmin = kwargs['tmin']
        if 'delta_t' in kwargs:
            delta_t = kwargs['delta_t']
        if 'immersion_time' in kwargs:
            immersion_time = kwargs['immersion_time']
            do_immersion = True
        t_i = immersion_time + delta_t*(2*np.random.random(loop) - 1)
        if 'emersion_time' in kwargs:
            emersion_time = kwargs['emersion_time']
            do_emersion = True
        t_e = emersion_time + delta_t*(2*np.random.random(loop) - 1)
        mask = (self.time >= tmin) & (self.time <= tmax)
        mask_sigma = (((self.time >= tmin) & (self.time < immersion_time - self.exptime)) +
                      ((self.time > emersion_time + self.exptime) & (self.time <= tmax)))
        sigma = kwargs.get('sigma', self.flux[mask_sigma].std(ddof=1))
        if 'dopacity' in kwargs:
            delta_opacity = kwargs['dopacity']
            do_opacity = True
        opas = opacity + delta_opacity*(2*np.random.random(loop) - 1)
        opas[opas > 1.], opas[opas < 0.] = 1.0, 0.0
        flux_min = 0
        flux_max = 1
        if 'flux_min' in kwargs:
            flux_min = kwargs['flux_min']
        if 'flux_max' in kwargs:
            flux_max = kwargs['flux_max']

        tflag = np.zeros(loop)
        tflag[t_i > t_e] = t_i[t_i > t_e]
        t_i[t_i > t_e] = t_e[t_i > t_e]
        t_e[t_i > t_e] = tflag[t_i > t_e]
        chi2 = 999999*np.ones(loop)
        for i in range(loop):
            model_test = self.__occ_model(t_i[i], t_e[i], opas[i], mask, flux_min=flux_min, flux_max=flux_max)
            chi2[i] = np.sum((self.flux[mask] - model_test)**2)/(sigma**2)
        kkargs = {}
        if do_immersion:
            kkargs['immersion'] = t_i
        if do_emersion:
            kkargs['emersion'] = t_e
        if do_opacity:
            kkargs['opacity'] = opas
        chisquare = ChiSquare(chi2, len(self.flux[mask]), **kkargs)
        onesigma = chisquare.get_nsigma(1)
        if 'immersion' in onesigma:
            self._immersion = self.tref + onesigma['immersion'][0]*u.s
            self.immersion_err = onesigma['immersion'][1]
            immersion_time = onesigma['immersion'][0]
        else:
            try:
                immersion_time = (self._immersion.jd - self.tref.jd)*u.d.to('s')
            except:
                pass
        if 'emersion' in onesigma:
            self._emersion = self.tref + onesigma['emersion'][0]*u.s
            self.emersion_err = onesigma['emersion'][1]
            emersion_time = onesigma['emersion'][0]
        else:
            try:
                emersion_time = (self._emersion.jd - self.tref.jd)*u.d.to('s')
            except:
                pass
        if 'opacity' in onesigma:
            opacity = onesigma['opacity'][0]
        # Run occ_model() to save best parameters in the Object.
        self.occ_model(immersion_time, emersion_time, opacity, np.repeat(True, len(self.flux)), flux_min=flux_min, flux_max=flux_max)
        self.lc_sigma = sigma
        self.chisquare = chisquare
        self.opacity = opacity
        return chisquare

    def plot_lc(self):
        """ Plots the light curve
        """
        if any(self.flux):
            pl.close()
            pl.plot(self.time, self.flux, 'k.-', label='Obs.', zorder=0)
            if any(self.model):
                pl.plot(self.time, self.model, 'r-', label='Model', zorder=2)
                pl.scatter(self.time, self.model, s=50, facecolors='none', edgecolors='r', zorder=3)
            pl.tight_layout()
            pl.xlabel('Time [seconds]', fontsize=20)
            pl.ylabel('Relative Flux', fontsize=20)
            pl.legend()
        else:
            raise ValueError('Plotting the light curve is only possible when the '
                             'Object LightCurve is instatiated with time and flux')
        return

    def plot_model(self):
        """ Plots the modelled light curve
        """
        if all(self.time_model):
            pl.plot(self.time_model, self.model_geometric, 'c-', label='Geometric', zorder=1)
            pl.plot(self.time_model, self.model_fresnel, 'b-', label='Fresnel', zorder=1)
            pl.plot(self.time_model, self.model_star, 'g-', label='Star diam.', zorder=1)
            pl.tight_layout()
            pl.xlabel('Time [seconds]', fontsize=20)
            pl.ylabel('Relative Flux', fontsize=20)
            pl.legend()
        else:
            raise ValueError('Plotting the model light curve is only possible after the model '
                             '[LightCurve.occ_model()] or the fit [LightCurve.occ_lcfit()]')
        return

    def to_log(self, namefile=None):
        """ Saves the light curve log to a file

        Parameters:
            namefile (str): Filename to save the log
        """
        if namefile is None:
            namefile = self.name.replace(' ', '_')+'.log'
        f = open(namefile, 'w')
        f.write(self.__str__())
        f.close()

    def to_file(self, namefile=None):
        """ Saves the light curve to a file

        Parameters:
            namefile (str): Filename to save the data
        """
        # Observational data
        if namefile is None:
            folder = ''
            file = self.name.replace(' ', '_')+'.dat'
        else:
            folder = os.path.dirname(namefile)
            file = os.path.basename(namefile)
        data = np.array([(self.time*u.s + self.tref).jd, self.time, self.flux, self.model, self.flux-self.model])
        colunm_names = ['Time JD', 'Time relative to {} UTC in seconds'.format(self.tref.iso),
                        'Observational Flux', 'Modelled Flux', 'Residual O-C']
        np.savetxt(os.path.join(folder, file), data.T, fmt='%11.8f')
        f = open(os.path.join(folder, file) + '.label', 'w')
        for i, name in enumerate(colunm_names):
            f.write('Column {}: {}\n'.format(i+1, name))
        f.close()
        # Complete Model
        if all(self.time_model):
            data_model = np.array([(self.time_model*u.s + self.tref).jd, self.time_model, self.model_geometric,
                                   self.model_fresnel, self.model_star])
            colunm_names_model = ['Model time JD', 'Model time relative to {} UTC in seconds'.format(self.tref.iso),
                                  'Geometric Model', 'Model with Fresnel diffraction', 'Model with star diameter']
            np.savetxt(os.path.join(folder, 'model_'+file), data_model.T, fmt='%11.8f')
            f = open(os.path.join(folder, 'model_'+file)+'.label', 'w')
            for i, name in enumerate(colunm_names_model):
                f.write('Column {}: {}\n'.format(i+1, name))
            f.close()

    def occ_detect(self, maximum_duration=None, dur_step=None, snr_limit=None,
                   n_detections=None, plot=False):
        """ Detects automatically the occultation event in the light curve
            (detects a 'square well transit')

        Parameters:
        (All parameters are optional)
            maximum_duration (float): Maximum duration of the occultation event. Default: light curve's time span
            dur_step (float): Step size to sweep occultation duration event. Default=1/2 of sampling
            snr_limit (float): Minimum occultation SNR. Default=none
            n_detections (int): Number of detections regardless the SNR.
                n_detections is superseded by snr_limit. Default=1
            plot (boolean): True if output plots are desired.

        Returns:
            OrderedDict = An ordered dictionary of :attr:`name`::attr:`value` pairs for each Parameter.

        Examples:
            >>> lc = sora.LightCurve(time=time, flux=flux, exptime=0.0, name='lc_example')
            >>> params = lc.occ_detect()
            >>> params
            {'rank': 1,
            'occultation_duration': 40.1384063065052,
            'central_time': 7916.773870512843,
            'immersion_time': 7896.7046673595905,
            'emersion_time': 7936.843073666096,
            'time_err': 0.05011036992073059,
            'depth': 0.8663887801707082,
            'depth_err': 0.10986223384336465,
            'baseline': 0.9110181732552853,
            'baseline_err': 0.19045768512595365,
            'snr': 7.886138392251848,
            'occ_mask': array([False, False, False, ..., False, False, False])}
        """

        if not hasattr(self, 'flux'):
            raise ValueError('time and flux must be instantiated to use ',
                             'occ_detect function.')

        # duration of the light curve
        time_span = self.time[-1]-self.time[0]
        if maximum_duration and (maximum_duration > time_span):
            warnings.warn('Occultation duration (maximum_duration={0}) ',
                          'exceeds the time series lenght ({1:0.5f}).',
                          ' maximum_duration reset to the time series lenght.'
                          .format(maximum_duration, time_span))
            maximum_duration = time_span
        if not maximum_duration:
            maximum_duration = time_span

        if not dur_step:
            dur_step = self.cycle/2

        if dur_step < self.cycle/2:
            warnings.warn('The given dur_step is oversampled by a factor ',
                          'of {0:0.1f} and has been reset to half a cycle ',
                          '({1:0.4f}).'
                          .format((self.cycle/2.)/dur_step, self.cycle/2.))
            dur_step = self.cycle/2

        duration_grid = np.arange(dur_step, maximum_duration, dur_step)
        # initial occultation mask (all data points)
        mask = np.ones(len(self.time), dtype=bool)
        # inital detection rank
        rank = 1

        if snr_limit:
            # minimum SNR accepted in a detection for multiple search
            snr_value = snr_limit+1
            occ0 = self.__run_bls(time_span, duration_grid)
            mask *= ~occ0['occ_mask']
            while (snr_value > snr_limit):
                rank += 1
                occ1 = self.__run_bls(time_span, duration_grid, mask=mask,
                                      rank=rank)
                if occ1['snr'] > snr_limit:
                    snr_value = occ1['snr']
                    mask *= ~occ1['occ_mask']
                    occ0 = self.__summarize_bls(occ0, occ1)
                else:
                    snr_value = snr_limit

            if plot:
                self.__plot_occ_detect(occ0)
            return occ0
        elif n_detections:
            # search the n best fits
            occ0 = self.__run_bls(time_span, duration_grid)
            mask *= ~occ0['occ_mask']
            for i in range(n_detections-1):
                rank += 1
                occ1 = self.__run_bls(time_span, duration_grid, mask=mask,
                                      rank=rank)
                snr_value = occ1['snr']
                mask *= ~occ1['occ_mask']
                occ0 = self.__summarize_bls(occ0, occ1)

            if plot:
                self.__plot_occ_detect(occ0)
            return occ0
        else:
            # search only the first best fit
            occ0 = self.__run_bls(time_span, duration_grid)

            if plot:
                self.__plot_occ_detect(occ0)

            return occ0

    def __plot_occ_detect(self, occ):
        n = np.size(occ['rank'])
        if n > 1:
            # case for multiple detections
            pl.plot(self.time, self.flux, 'k.-')
            mask = np.zeros(len(self.time), dtype=bool)
            for i in range(n):
                trues = np.sum(occ['occ_mask'][i])
                pl.plot(self.time[occ['occ_mask'][i]], np.repeat(np.mean(self.flux[occ['occ_mask'][i]]), trues),
                        '.', label='Rank: '+str(i+1))
                mask += occ['occ_mask'][i]

            falses = list(mask).count(False)
            pl.plot(self.time[~mask], np.repeat(np.mean(self.flux[~mask]), falses), 'r.', label='Baseline')
            pl.xlabel('Time [seconds]')
            pl.ylabel('Relative Flux')
            pl.legend()

        else:
            # case for single occultation
            trues = list(occ['occ_mask']).count(True)
            falses = list(occ['occ_mask']).count(False)
            pl.plot(self.time, self.flux, 'k.-')
            pl.plot(self.time[occ['occ_mask']], np.repeat(np.mean(self.flux[occ['occ_mask']]), trues),
                    '.', label='Occultation')
            pl.plot(self.time[~occ['occ_mask']], np.repeat(np.mean(self.flux[~occ['occ_mask']]), falses),
                    'r.', label='Baseline')
            pl.xlabel('Time [seconds]')
            pl.ylabel('Relative Flux')
            pl.legend()

    def __run_bls(self, per_grid, dur_grid, mask=None, rank=None):
        """ Private function to find the best box fit suitable to the data
        """

        # object with no occultation mask
        mskmodel = BoxLeastSquares(self.time, self.flux, dy=self.dflux)
        # if there is no dflux array, reset it to None in case of
        # using a mask
        if self.dflux is None:
            dfluxmask = None
        else:
            dfluxmask = self.dflux[mask]

        # object with occultation mask
        if np.sum(mask):
            model = BoxLeastSquares(self.time[mask], self.flux[mask],
                                    dy=dfluxmask)
        else:
            model = mskmodel

        r = model.power(per_grid, dur_grid, objective='snr', method='fast')
        # statistics of the BLS fit
        stats = model.compute_stats(r.period, r.duration, r.transit_time)
        # occultation mask of the event with respect to all data
        occ_mask = mskmodel.transit_mask(self.time, r.period, r.duration,
                                         r.transit_time)
        # parameters computation for clarity purposes
        occultation_duration = r.duration[0]
        central_time = stats['transit_times'][0]
        immersion_time = stats['transit_times'][0] - r.duration[0]/2
        emersion_time = stats['transit_times'][0] + r.duration[0]/2
        time_err = np.median(self.time[1:-1]-self.time[0:-2])/2
        depth = np.mean(self.flux[~occ_mask])-np.mean(self.flux[occ_mask])
        depth_err = np.std(self.flux[occ_mask], ddof=1)
        baseline = np.mean(self.flux[~occ_mask])
        baseline_err = np.std(self.flux[~occ_mask], ddof=1)
        # If there is only one measurement during the occultation it will
        # use the baseline_err to compute SNR, otherwise it will use depth_err
        if np.sum(occ_mask) < 2:
            snr = depth/baseline_err
        else:
            snr = depth/depth_err

        # define rank
        if rank:
            rank = rank
        else:
            rank = 1

        return {'rank': rank,
                'occultation_duration': occultation_duration,
                'central_time': central_time,
                'immersion_time': immersion_time,
                'emersion_time': emersion_time,
                'time_err': time_err,
                'depth': depth, 'depth_err': depth_err,
                'baseline': baseline, 'baseline_err': baseline_err,
                'snr': snr, 'occ_mask': occ_mask}

    def __summarize_bls(self, dict1, dict2):
        """ Private function to merge dictionaries returned by BLS and
            to keep values of common keys in list.
        """
        dict3 = {}
        for key, value in dict1.items():
            if key == 'occ_mask':
                sz = int(np.size(dict1[key])/np.size(dict2[key]))
                if sz > 1:
                    k = [None]*(sz+1)
                    for i in range(sz):
                        k[i] = dict1[key][i]
                    k[i+1] = dict2[key]
                    dict3[key] = k
                else:
                    dict3[key] = [dict1[key], dict2[key]]
            else:
                dict3[key] = np.append(dict1[key], dict2[key])
        return dict3

    def __bar_fresnel(self, X, X01, X02, fresnel_scale, opacity):
        """ Returns the modelled light curve considering fresnel difraction.

        Parameters:
            X (array): Array with time values converted in km using the event velocity.
            X01 (int, float): Immersion time converted in km using the event velocity.
            X02 (int, float): Emersion time converted in km using the event velocity.
            fresnel_scale (int, float): Fresnel scale, in km.
            opacity (int, float): Opacity. Opaque = 1.0, transparent = 0.0

        Returns:
            flux_fresnel (array): the light curve with fresnel diffraction
        """
        # Converting from km to units of fresnel scale
        x = X/fresnel_scale
        x01 = X01/fresnel_scale
        x02 = X02/fresnel_scale
        # Fresnel difraction parameters
        x1 = x - x01
        x2 = x - x02
        s1, c1 = scsp.fresnel(x1)
        s2, c2 = scsp.fresnel(x2)
        cc = c1 - c2
        ss = s1 - s2
        r_ampli = - (cc+ss)*(opacity/2.)
        i_ampli = (cc-ss)*(opacity/2.)
        # Determining the flux considering fresnel difraction
        flux_fresnel = (1.0 + r_ampli)**2 + (i_ampli)**2
        return flux_fresnel

    def __occ_model(self, immersion_time, emersion_time, opacity, mask, npt_star=12,
                    time_resolution_factor=10, flux_min=0.0, flux_max=1.0):
        """ Private function returns the modelled light curve considering fresnel difraction,
            star diameter and intrumental response, intended for fitting inside the self.occ_lcfit().

        Parameters:
            immersion_time (int, float): Immersion time, in seconds.
            emersion_time (int, float): Emersion time, in seconds.
            opacity (int, float): Opacity. Opaque = 1.0, transparent = 0.0.
            mask (array with Booleans): Mask with True values to be computed
            npt_star (int): Number of subdivisions for computing the star size's effects. Default=12
            time_resolution_factor (int,float): Steps for fresnel scale used for modelling the light curve.
                Default=10*fresnel scale.
            flux_min (int,float): Bottom flux (only object). Default=0.0
            flux_max (int,float): Base flux (object plus star). Default=1.0

        Returns:
            flux_inst (array): Modelled Instrumental light flux.
        """
        # Computing the fresnel scale
        lamb = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist = self.dist*u.au.to('km')
        vel = np.absolute(self.vel)
        time_obs = self.time[mask]
        fresnel_scale_1 = calc_fresnel(dist, lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist, lamb+dlamb/2.0)
        fresnel_scale = (fresnel_scale_1 + fresnel_scale_2)/2.0
        time_resolution = (np.min([fresnel_scale/vel, self.exptime]))/time_resolution_factor
        self.model_resolution = time_resolution

        # Creating a high resolution curve to compute fresnel difraction, stellar diameter and instrumental integration
        time_model = np.arange(time_obs.min()-5*self.exptime, time_obs.max()+5*self.exptime, time_resolution)

        # Changing X: time (s) to distances in the sky plane (km), considering the tangential velocity (vel in km/s)
        x = time_model*vel
        x01 = immersion_time*vel
        x02 = emersion_time*vel

        # Computing fresnel diffraction for the case where the star size is negligenciable
        flux_fresnel_1 = self.__bar_fresnel(x, x01, x02, fresnel_scale_1, opacity)
        flux_fresnel_2 = self.__bar_fresnel(x, x01, x02, fresnel_scale_2, opacity)
        flux_fresnel = (flux_fresnel_1 + flux_fresnel_2)/2.
        flux_star = flux_fresnel.copy()
        if (self.d_star > 0):
            # Computing fresnel diffraction for the case where the star size is not negligenciable
            resolucao = self.d_star/npt_star
            flux_star_1 = np.zeros(len(time_model))
            flux_star_2 = np.zeros(len(time_model))
            # Computing stellar diameter only near the immersion or emersion times
            star_diam = (np.absolute(x - x01) < 3*self.d_star) + (np.absolute(x - x02) < 3*self.d_star)
            p = np.arange(-npt_star, npt_star)*resolucao
            coeff = np.sqrt(np.absolute(self.d_star**2 - p**2))
            for ii in np.where(star_diam == True)[0]:
                xx = x[ii] + p
                flux1 = self.__bar_fresnel(xx, x01, x02, fresnel_scale_1, opacity)
                flux2 = self.__bar_fresnel(xx, x01, x02, fresnel_scale_2, opacity)
                flux_star_1[ii] = np.sum(coeff*flux1)/coeff.sum()
                flux_star_2[ii] = np.sum(coeff*flux2)/coeff.sum()
                flux_star[ii] = (flux_star_1[ii] + flux_star_2[ii])/2.
        flux_inst = np.zeros(len(time_obs))
        for i in range(len(time_obs)):
            event_model = (time_model > time_obs[i]-self.exptime/2.) & (time_model < time_obs[i]+self.exptime/2.)
            flux_inst[i] = (flux_star[event_model]).mean()
        return flux_inst*(flux_max - flux_min) + flux_min

    def __str__(self):
        """ String representation of the LightCurve Object
        """
        output = 'Light curve name: {}\n'.format(self.name)
        try:
            output += ('Initial time: {} UTC\n'
                       'End time:     {} UTC\n'
                       'Duration:     {:.3f} minutes\n'.format(
                           self.initial_time.iso, self.end_time.iso,
                           (self.end_time - self.initial_time).value*u.d.to('min'))
                       )
        except:
            pass
        output += 'Time offset:  {:.3f} seconds\n\n'.format(self.dt)
        try:
            output += 'Exposure time:    {:.4f} seconds\n'.format(self.exptime)
            output += 'Cycle time:       {:.4f} seconds\n'.format(self.cycle)
            output += 'Num. data points: {}\n\n'.format(len(self.time))
        except:
            output += 'Object LightCurve was not instantiated with time and flux.\n\n'
        try:
            output += ('Bandpass:            {:.3f} +/- {:.3f} microns\n'
                       'Used shadow velocity: {:.3f} km/s\n'
                       'Fresnel scale:        {:.3f} seconds or {:.2f} km\n'
                       'Stellar size effect:  {:.3f} seconds or {:.2f} km\n'.format(
                           self.lambda_0, self.delta_lambda, self.vel, self.fresnel_scale/self.vel,
                           self.fresnel_scale, self.d_star/self.vel, self.d_star)
                       )
        except:
            output += '\nThere is no occultation associated with this light curve.\n'
        try:
            output += ('Inst. response:       {:.3f} seconds or {:.2f} km\n'
                       'Dead time effect:     {:.3f} seconds or {:.2f} km\n'
                       'Model resolution:     {:.3f} seconds or {:.2f} km\n'
                       'Modelled baseflux:    {:.3f}\n'
                       'Modelled bottomflux:  {:.3f}\n'
                       'Light curve sigma:    {:.3f}\n\n'.format(
                           self.exptime, self.exptime*self.vel, self.cycle-self.exptime,
                           (self.cycle-self.exptime)*self.vel, self.model_resolution,
                           self.model_resolution*self.vel, self.baseflux, self.bottomflux,
                           self.lc_sigma)
                       )
        except:
            output += '\nObject LightCurve model was not fitted.\n\n'
        try:
            output += ('Immersion time: {} UTC +/- {:.3f} seconds\n'
                       'Emersion time:  {} UTC +/- {:.3f} seconds\n\n'.format(
                           self.immersion.iso, self.immersion_err,
                           self.emersion.iso, self.emersion_err)
                       )
        except:
            output += 'Immersion and emersion times were not fitted or instantiated.\n\n'

        try:
            output += 'Monte Carlo chi square fit.\n\n' + self.chisquare.__str__() + '\n'
        except:
            pass
        return output

    def __del__(self):
        """ When this object is deleted, it removes the name from the Class name list.
        """
        try:
            self.__names.remove(self.__name)
        except:
            pass



def check_kwargs(input_kwargs, allowed_kwargs, raise_error=True):
    """ Tests if the input kwargs are allowed
    
    Parameters:
        input_kwargs: Dictionary or list with the input values
        allowed_kwargs: list with the allowed keys
        raise_error: if True, it will raises an error,
            if False, it returns the keys not allowed.
    """
    not_allowed = [i for i in input_kwargs if i not in allowed_kwargs]
    if raise_error:
        if len(not_allowed) > 0:
            allowed_kwargs.sort()
            raise TypeError("function got an unexpected keyword argument {}\n"
                            "Available kwargs are: {}".format(not_allowed, allowed_kwargs))
    else:
        return not_allowed




from astropy.coordinates import SkyCoord, SphericalCosLatDifferential, Distance
from astropy.coordinates import get_sun, SphericalRepresentation, SkyOffsetFrame, ICRS
from astropy.time import Time
from astropy.table import Table
import astropy.units as u
from astroquery.vizier import Vizier
import warnings
import numpy as np
from sora.config import test_attr, input_tests


warnings.simplefilter('always', UserWarning)


def search_star(**kwargs):
    """ Searches position on VizieR and returns a catalogue.

    Parameters:
        coord (str, SkyCoord): Coordinate to perform the search.
        code (str): Gaia Source_id of the star
        columns (list): List of strings with the name of the columns to retrieve.
        radius (int, float, unit.quantity): Radius to search around coordinate
        catalog (str): VizieR catalogue to search.

    Returns:
        catalogue(astropy.Table): An astropy Table with the catalogue informations.
    """
    input_tests.check_kwargs(kwargs, allowed_kwargs=['catalog', 'code', 'columns', 'coord', 'log', 'radius'])
    row_limit = 100
    if 'log' in kwargs and kwargs['log']:
        print('Downloading star parameters from {}'.format(kwargs['catalog']))
    vquery = Vizier(columns=kwargs['columns'], row_limit=row_limit, timeout=600)
    if 'code' in kwargs:
        catalogue = vquery.query_constraints(catalog=kwargs['catalog'], Source=kwargs['code'], cache=False)
    elif 'coord' in kwargs:
        catalogue = vquery.query_region(kwargs['coord'], radius=kwargs['radius'], catalog=kwargs['catalog'], cache=False)
    else:
        raise ValueError('At least a code or coord should be given as input')
    return catalogue


def van_belle(magB=None, magV=None, magK=None):
    """ Determines the diameter of a star in mas using equations from van Belle (1999)
        -- Publi. Astron. Soc. Pacific 111, 1515-1523:

    Parameters:
        magB: The magnitude B of the star
        magV: The magnitude V of the star
        magK: The magnitude K of the star
        If any of those values is 'None', 'nan' or higher than 49, it is not considered.
    """
    if magB is None or np.isnan(magB) or magB > 49:
        magB = np.nan
    if magV is None or np.isnan(magV) or magV > 49:
        magV = np.nan
    if magK is None or np.isnan(magK) or magK > 49:
        magK = np.nan

    def calc_diameter(a1, a2, mag):
        return 10**(a1 + a2*(mag - magK) - 0.2*mag)

    params = {'sg': {'B': [0.648, 0.220], 'V': [0.669, 0.223]},
              'ms': {'B': [0.500, 0.290], 'V': [0.500, 0.264]},
              'vs': {'B': [0.840, 0.211], 'V': [0.789, 0.218]}}

    mag = np.array([magB, magV])
    diameter = {}
    for st in ['sg', 'ms', 'vs']:
        diameter_s = {}
        for i, m in enumerate(['B', 'V']):
            diam = calc_diameter(*params[st][m], mag[i])
            if not np.isnan(diam):
                diameter_s[m] = calc_diameter(*params[st][m], mag[i])*u.mas
        if diameter_s:
            diameter[st] = diameter_s
    return diameter


def kervella(magB=None, magV=None, magK=None):
    """ Determines the diameter of a star in mas using equations from Kervella et. al (2004)
        -- A&A Vol. 426, No.  1:

    Parameters:
        magB: The magnitude B of the star
        magV: The magnitude V of the star
        magK: The magnitudes K of the star
        If any of those values is 'None', 'nan' or higher than 49, it is not considered.
    """
    if magB is None or np.isnan(magB) or magB > 49:
        magB = np.nan
    if magV is None or np.isnan(magV) or magV > 49:
        magV = np.nan
    if magK is None or np.isnan(magK) or magK > 49:
        magK = np.nan
    const1 = np.array([0.0755, 0.0535])
    const2 = np.array([0.5170, 0.5159])
    mag = np.array([magV, magB])
    vals = 10**(const1*(mag-magK)+const2-0.2*magK)
    diam = {}
    if not np.isnan(vals[0]):
        diam['V'] = vals[0]*u.mas
    if not np.isnan(vals[1]):
        diam['B'] = vals[1]*u.mas
    return diam


class Star():
    def __init__(self, **kwargs):
        """ Defines a star

        Parameters:
            code (str): Gaia-DR2 Source code for searching in VizieR.
            coord (str, SkyCoord): if code is not given, coord nust have the coordinates
                RA and DEC of the star to search in VizieR: 'hh mm ss.ss +dd mm ss.ss'
            nomad (bool): If true, it tries to download the magnitudes from NOMAD catalogue.
            log (bool): If true, it prints the downloaded information.
            local (bool): If true, it uses the given coordinate in 'coord' as final coordinate.
        """
        self.__local = False
        self.mag = {}
        self.errors = {}
        self.__log = True
        input_tests.check_kwargs(kwargs, allowed_kwargs=['code', 'coord', 'local', 'log', 'nomad'])
        if 'log' in kwargs:
            self.__log = kwargs['log']
        if 'local' in kwargs:
            self.__local = test_attr(kwargs['local'], bool, 'local')
        if not any(i in kwargs for i in ['coord', 'code']):
            raise KeyError('Input values must have either the coordinate of the star'
                           ' or the Gaia code for online search')
        if 'coord' in kwargs:
            self.coord = SkyCoord(kwargs['coord'], unit=('hourangle', 'deg'))
        else:
            self.__local = False
        if 'code' in kwargs:
            self.code = test_attr(kwargs['code'], str, 'code')
        if not self.__local:
            self.__searchgaia()
        nomad = kwargs.get('nomad', True)
        if nomad:
            self.__getcolors()

    def set_magnitude(self, **kwargs):
        """ Sets the magnitudes of a star.

        Parameters:
            (band name)=(float): The star magnitude for given band. The band name can be any string the user wants.

        Examples:
            set_magnitude(G=10)
            set_magnitude(K=15)
            set_magnitude(newband=6)
        """
        for key in kwargs:
            mag = test_attr(kwargs[key], float, key)
            if key in self.mag:
                warnings.warn('{0} mag already defined. {0}={1} will be replaced by {0}={2}'.format(
                    key, self.mag[key], mag))
            self.mag[key] = mag

    def set_diameter(self, diameter):
        """ Sets an user diameter for the star, in mas.

        Parameters:
            diameter (int,float): sets the user diameter of the star, in mas
        """
        self.diameter_user = diameter*u.mas

    def van_belle(self):
        """ Determines the diameter of a star in mas using equations from van Belle (1999)
            -- Publi. Astron. Soc. Pacific 111, 1515-1523:
        """
        return van_belle(self.mag.get('B'), self.mag.get('V'), self.mag.get('K'))

    def kervella(self):
        """ Determines the diameter of a star in mas using equations from Kervella et. al (2004)
            -- A&A Vol.  426, No.  1:
        """
        return kervella(self.mag.get('B'), self.mag.get('V'), self.mag.get('K'))

    def apparent_diameter(self, distance, mode='auto', band='V', star_type='sg', log=True):
        """ Calculates the apparent diameter of the star at a given distance

        Parameters:
            distance (int, float): Object geocentric distance, in AU
            mode (str): The mode to calculate the apparent diameter
                'user': calculates using user given diameter
                'gaia': calculates using diameter obtained from Gaia
                'kervella': calculates using Kervella equations
                'van_belle': calculates using van Belle equations
                'auto' (default): tries all the above methods until it is able to calculate diameter.
                    The order of try is the same as shown above (user, Gaia, Kervella, Van Belle).
            'band' (str): The band filter to calculate the diameter.
                If mode is 'kervella' or 'van_belle', the filter must be given, 'B' or 'V'.
                If mode 'auto', 'V' is selected.
            'star_type' (str): type of star to calculate the diameter.
                If mode is 'van_belle', the star type must be given.
                If mode is 'auto', type = 'sg'.
                Types can be:
                    - 'sg' for 'Super Giant'
                    - 'ms' for 'Main Sequence'
                    - 'vs' for 'Variable Star'
            'log' (bool): If True, it prints the mode used by 'auto'.
        """
        try:
            distance = distance.to(u.km)
        except:
            distance = distance*u.AU

        if mode in ['user', 'auto']:
            try:
                diam = distance*np.tan(self.diameter_user)
                if log:
                    print('Calculating apparent diameter from user defined diameter')
                return diam.to(u.km)
            except:
                pass

        if mode == 'user':
            raise ValueError('User diameter must be informed.')

        if mode in ['gaia', 'auto']:
            try:
                diam = distance*np.tan(self.diameter_gaia)
                if log:
                    print('Apparent diameter using Gaia')
                return diam.to(u.km)
            except:
                pass

        if mode == 'gaia':
            raise ValueError('It is not possible to calculate star diameter from Gaia.')

        if band not in ['B', 'V']:
            raise KeyError('band must be informed as "B", or "V"')

        if mode in ['kervella', 'auto']:
            diam_kerv = self.kervella().get(band)
            if diam_kerv is None:
                raise ValueError('Diameter could not be calculated for given band')
            if log:
                print('Apparent diameter using Kervella et al. (2004)')
            diam = distance*np.tan(diam_kerv)
            return diam.to(u.km)

        if star_type not in ['sg', 'ms', 'vs']:
            raise KeyError('star_type must be informed as "sg", "ms" or "vs"')

        if mode in ['van_belle', 'auto']:
            diam_van = self.van_belle().get(star_type)
            if diam_van is None:
                raise ValueError('Diameter could not be calculated using Van Belle')
            diam_van = diam_van.get(band)
            if diam_van is None:
                raise ValueError('Diameter could not be calculated for given band')
            if log:
                print('Apparent diameter using van Belle (1999)')
            diam = distance*np.tan(diam_van)
            return diam.to(u.km)

        raise AttributeError("Star apparent diameter could not be calculated. ",
                             "Please define star diameter or B,V,K magnitudes.")

    def __searchgaia(self):
        """ Searches for the star position in the Gaia catalogue and save informations
        """
        if hasattr(self, 'code'):
            catalogue = search_star(code=self.code, columns=['**'], catalog='I/345/gaia2', log=self.__log)
        else:
            catalogue = search_star(coord=self.coord, columns=['**'], radius=2*u.arcsec,
                                    catalog='I/345/gaia2', log=self.__log)
        if len(catalogue) == 0:
            raise ValueError('No star was found. It does not exist or VizieR is out.')
        catalogue = catalogue[0]
        if len(catalogue) > 1:
            if self.__log:
                print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
                print('The list below is sorted by distance. Please select the correct star')
            gmag = catalogue['Gmag']
            tstars = SkyCoord(catalogue['RA_ICRS'], catalogue['DE_ICRS'])
            sep = tstars.separation(self.coord)
            k = sep.argsort()
            while True:
                t = Table()
                t['num'] = np.arange(len(tstars))+1
                t['dist(")'] = sep[k].arcsec
                t['dist(")'].format = '6.3f'
                t['Gmag'] = gmag[k].quantity.value
                t['Gmag'].format = '6.3f'
                t['RA___ICRS___DEC'] = tstars[k].to_string('hmsdms', precision=4)
                t.pprint_all()
                print('  0: Cancel')
                choice = int(input('Choose the corresponding number of the correct star: '))
                if choice in np.arange(len(k)+1):
                    break
                print('{} is not a valid choice. Please select the correct star'.format(choice))
            if choice == 0:
                raise ValueError('It was not possible to define a star')
            catalogue = catalogue[[k[choice-1]]]
        self.code = catalogue['Source'][0]
        ra = catalogue['RA_ICRS']
        dec = catalogue['DE_ICRS']
        pmra = catalogue['pmRA']
        pmde = catalogue['pmDE']
        epoch = Time(catalogue['Epoch'].quantity, format='jyear')
        try:
            distance = Distance(parallax=catalogue['Plx'].quantity, allow_negative=False)
            self.coord = SkyCoord(ra, dec, distance=distance, pm_ra_cosdec=pmra, pm_dec=pmde, obstime=epoch[0])[0]
        except:
            self.coord = SkyCoord(ra, dec, pm_ra_cosdec=pmra, pm_dec=pmde, obstime=epoch[0])[0]
        self.set_magnitude(G=catalogue['Gmag'][0])
        self.errors['RA'] = catalogue['e_RA_ICRS'][0]*u.mas
        self.errors['DEC'] = catalogue['e_DE_ICRS'][0]*u.mas
        self.errors['pmRA'] = catalogue['e_pmRA'][0]*(u.mas/u.yr)
        self.errors['pmDEC'] = catalogue['e_pmDE'][0]*(u.mas/u.yr)
        rad = catalogue['Rad'][0]
        self.meta_gaia = {c: catalogue[c][0] for c in catalogue.columns}
        if np.ma.core.is_masked(rad) or np.ma.core.is_masked(catalogue['Plx'][0]):
            if self.__log:
                warnings.warn('Gaia catalogue does not have the star radius.')
        else:
            self.diameter_gaia = 2*np.arctan((rad*u.solRad)/distance[0]).to(u.mas)
        if self.__log:
            print('1 Gaia-DR2 star found G={}'.format(catalogue['Gmag'][0]))
            print('star coordinate at J{}: RA={} +/- {}, DEC={} +/- {}'.format(self.coord.obstime.jyear,
                  self.coord.ra.to_string(u.hourangle, sep='hms', precision=5), self.errors['RA'],
                  self.coord.dec.to_string(u.deg, sep='dms', precision=4), self.errors['DEC']))

    def __getcolors(self):
        """ Searches for the B,V,K magnitudes of the star in the NOMAD catalogue on VizieR
        """
        columns = ['RAJ2000', 'DEJ2000', 'Bmag', 'Vmag', 'Rmag', 'Jmag', 'Hmag', 'Kmag']
        catalogue = search_star(coord=self.coord, columns=columns, radius=2*u.arcsec,
                                catalog='I/297/out', log=self.__log)
        if len(catalogue) == 0:
            if self.__log:
                warnings.warn('No star was found on NOMAD that matches the star')
            return
        catalogue = catalogue[0]
        if len(catalogue) > 1:
            print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
            print('The list below is sorted by distance. Please select the correct star')
            print('Star G mag: {}'.format(self.mag['G']))
            tstars = SkyCoord(catalogue['RAJ2000'], catalogue['DEJ2000'])
            sep = tstars.separation(self.coord)
            k = sep.argsort()
            while True:
                t = Table()
                t['num'] = np.arange(len(tstars))+1
                t['dist(")'] = sep[k].arcsec
                t['dist(")'].format = '6.3f'
                t['Bmag'] = catalogue['Bmag'][k].quantity.value
                t['Vmag'] = catalogue['Vmag'][k].quantity.value
                t['Rmag'] = catalogue['Rmag'][k].quantity.value
                t['Jmag'] = catalogue['Jmag'][k].quantity.value
                t['Hmag'] = catalogue['Hmag'][k].quantity.value
                t['Kmag'] = catalogue['Kmag'][k].quantity.value
                t['Bmag'].format = t['Vmag'].format = t['Rmag'].format = '6.3f'
                t['Jmag'].format = t['Hmag'].format = t['Kmag'].format = '6.3f'
                t['RA___ICRS___DEC'] = tstars[k].to_string('hmsdms', precision=4)
                t.pprint_all()
                print('  0: Cancel')
                choice = int(input('Choose the corresponding number of the correct star: '))
                if choice in np.arange(len(k)+1):
                    break
                print('{} is not a valid choice. Please select the correct star'.format(choice))
            if choice == 0:
                warnings.warn('No magnitudes were obtained from NOMAD')
                return
            catalogue = catalogue[[k[choice-1]]]
        errors = []
        for mag in ['B', 'V', 'R', 'J', 'H', 'K']:
            name = mag + 'mag'
            if np.ma.core.is_masked(catalogue[name][0]):
                errors.append(mag)
                continue
            self.set_magnitude(**{mag: catalogue[name][0]})
        if len(errors) > 0 and self.__log:
            warnings.warn('Magnitudes in {} were not located in NOMAD'.format(errors))

    def geocentric(self, time):
        """ Calculates the position of the star, propagating the position using parallax and proper motion

        Parameters:
            time (float, Time): reference time to apply proper motion and calculate paralax.
        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        n_coord = self.barycentric(time)
        if self.coord.distance.unit.is_unity() or np.isnan(self.coord.distance):
            g_coord = n_coord
        else:
            sun = get_sun(time)
            g_coord = SkyCoord(*(n_coord.cartesian.xyz + sun.cartesian.xyz),
                               representation_type='cartesian')
            g_coord = g_coord.represent_as(SphericalRepresentation)
            g_coord = SkyCoord(g_coord.lon, g_coord.lat, g_coord.distance)

        if hasattr(self, 'offset'):
            star_frame = SkyOffsetFrame(origin=g_coord)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat, frame=star_frame)
            return new_pos.transform_to(ICRS)

        return g_coord

    def barycentric(self, time):
        """ Calculates the position of the star using proper motion

        Parameters:
            time (str, Time): reference time to apply proper motion.
        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        if np.isnan(self.coord.pm_dec):
            return self.coord
        if self.coord.distance.unit.is_unity():
            star_frame = SkyOffsetFrame(origin=self.coord)
            dt = time - self.coord.obstime
            new_pos = SkyCoord(lon=self.coord.pm_ra_cosdec*dt, lat=self.coord.pm_dec*dt, frame=star_frame)
            n_coord = new_pos.transform_to(ICRS)
        else:
            n_coord = self.coord.apply_space_motion(new_obstime=time)
        return n_coord

    def error_at(self, time):
        """ Estimates the star position error at a given time

        Parameters:
            time (str, Time): reference time to project star error.

        Returns:
            errors in RA* and DEC
        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        e_ra = self.errors['RA'] + self.errors['pmRA']*np.abs(time-self.coord.obstime)
        e_dec = self.errors['DEC'] + self.errors['pmDEC']*np.abs(time-self.coord.obstime)
        return e_ra, e_dec

    def add_offset(self, da_cosdec, ddec):
        """ Adds an offset to the star position

        Parameters:
            da_cosdec (int, float): offset in Delta_alpha_cos_delta in mas
            ddec (int, float): offset in Delta_delta in mas
        """
        dadc = test_attr(da_cosdec, float, 'da_cosdec')
        dd = test_attr(ddec, float, 'ddec')
        self.offset = SphericalCosLatDifferential(dadc*u.mas, dd*u.mas, 0.0*u.km)

    def __str__(self):
        """ String representation of the Star class
        """
        out = ''
        if hasattr(self, 'code'):
            out += 'Gaia-DR2 star Source ID: {}\n'.format(self.code)
        out += ('ICRS star coordinate at J{}:\n'
                'RA={} +/- {:.4f}, DEC={} +/- {:.4f}\n'
                'pmRA={:.3f} +/- {:.3f} mas/yr, pmDEC={:.3f} +/- {:.3f} mas/yr, Plx={:.4f} +/- {:.4f} mas\n\n'.format(
                    self.coord.obstime.jyear, self.coord.ra.to_string(u.hourangle, sep='hms', precision=5),
                    self.errors['RA'], self.coord.dec.to_string(u.deg, sep='dms', precision=4), self.errors['DEC'],
                    self.meta_gaia['pmRA'], self.meta_gaia['e_pmRA'], self.meta_gaia['e_pmDE'], self.meta_gaia['pmRA'],
                    self.meta_gaia['Plx'], self.meta_gaia['e_Plx']))
        if hasattr(self, 'offset'):
            out += 'Offset Apllied: d_alpha_cos_dec = {}, d_dec = {}\n'.format(
                self.offset.d_lon_coslat, self.offset.d_lat)
        out += 'Magnitudes:'
        mag_out = [' {}: {:6.3f}'.format(mag, self.mag[mag]) for mag in self.mag]
        out_mag = []
        for i, mag in enumerate(mag_out):
            if i % 6 == 0:
                out_mag.append([])
            out_mag[-1].append(mag)
        out += (',\n'+' '*11).join([','.join(out_i) for out_i in out_mag])
        out += '\n\n'
        if hasattr(self, 'diameter_gaia'):
            out += 'Apparent diameter: {:.4f}, Source: Gaia-DR2\n'.format(self.diameter_gaia)
        if hasattr(self, 'diameter_user'):
            out += 'Apparent diameter: {:.4f}, Source: User\n'.format(self.diameter_user)
        kerv = self.kervella()
        if kerv:
            out += 'Apparent diameter from Kervella et. al (2004):\n'
            out += '   ' + ','.join([' {}: {:.4f}'.format(k, v) for k, v in kerv.items()])
        vanb = self.van_belle()
        if vanb:
            out += '\nApparent diameter from van Belle (1999):'
            for key, value in vanb.items():
                out += '\n    {}:'.format(key)
                out += ','.join([' {}: {:.4f}'.format(k, v) for k, v in value.items()])
        return out



from astropy.coordinates import SkyCoord, EarthLocation, GCRS, AltAz
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.time import Time
import astropy.units as u
from astroquery.mpc import MPC
import numpy as np
from sora.config import test_attr, input_tests


def search_code_mpc():
    """ Reads the MPC Observer Database

    Returns:
        observatories (dict): A python dictionaty with all the sites as an Astropy EarthLocation object
    """
    obs = MPC.get_observatory_codes()
    observatories = {}
    for line in obs:
        code = line['Code']
        lon = line['Longitude']*u.deg
        rcphi = line['cos']*6378.137*u.km
        rsphi = line['sin']*6378.137*u.km
        name = line['Name']
        site = EarthLocation.from_geocentric(rcphi*np.cos(lon), rcphi*np.sin(lon), rsphi)
        observatories[code] = (name, site)
    return observatories


class Observer():
    __names = []

    def __init__(self, **kwargs):
        """ Defines the observer object

        Parameters:
            name (str): Name for the Observer. (required)
                Observer is uniquely defined (name must be different for each observer).
            code (str): The IAU code for SORA to search for its coordinates in MPC database
            site (EarthLocation): User provides an EarthLocation object.
            lon (str, float): The Longitude of the site in degrees.
                Positive to East. Range (0 to 360) or (-180 to +180)
                User can provide in degrees (float) or hexadecimal (string)
            lat (str, float): The Latitude of the site in degrees.
                Positive North. Range (+90 to -90)
                User can provide in degrees (float) or hexadecimal (string)
            height (int, float): The height of the site in meters above see level.

        Examples:
            User can provide one of the following to define an observer:
            - If user will use the MPC name for the site:
                Observer(code)
            - If user wants to use a different name from the MPC database:
                Observer(name, code)
            - If user wants to use an EarthLocation value:
                EarthLocation(lon, lat, height)
                Observer(name, site)
            - If user wants to give site coordinates directly:
                Observer(name, lon, lat, height)
        """
        input_tests.check_kwargs(kwargs, allowed_kwargs=['code', 'height', 'lat', 'lon', 'name', 'site'])
        if 'code' in kwargs:
            self.code = kwargs['code']
            try:
                name, self.site = search_code_mpc()[self.code]
                self.__name = kwargs.get('name', name)
            except:
                raise ValueError('code {} could not be located in MPC database'.format(self.code))
        elif all(i in kwargs for i in ['name', 'site']):
            self.__name = kwargs['name']
            self.site = test_attr(kwargs['site'], EarthLocation, 'site')
        elif all(i in kwargs for i in ['name', 'lon', 'lat', 'height']):
            self.__name = kwargs['name']
            self.site = EarthLocation(kwargs['lon'], kwargs['lat'], kwargs['height'])
        else:
            raise ValueError('Input parameters could not be determined')
        if self.__name in self.__names:
            raise ValueError("name {} already defined for another Observer object. "
                             "Please choose a different one.".format(self.__name))
        self.__names.append(self.__name)

    def get_ksi_eta(self, time, star):
        """ Calculates relative position to star in the orthographic projection.

        Parameters:
            time (str, Time): Reference time to calculate the position.
                It can be a string in the format "yyyy-mm-dd hh:mm:ss.s" or an astropy Time object
            star (str, SkyCoord): The coordinate of the star in the same reference frame as the ephemeris.
                It can be a string in the format "hh mm ss.s +dd mm ss.ss"
                or an astropy SkyCoord object.

        Returns:
            ksi, eta (float): on-sky orthographic projection of the observer relative to a star
                Ksi is in the North-South direction (North positive)
                Eta is in the East-West direction (East positive)
        """
        time = test_attr(time, Time, 'time')
        try:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        except:
            raise ValueError('star is not an astropy object or a string in the format "hh mm ss.s +dd mm ss.ss"')

        itrs = self.site.get_itrs(obstime=time)
        gcrs = itrs.transform_to(GCRS(obstime=time))
        rz = rotation_matrix(star.ra, 'z')
        ry = rotation_matrix(-star.dec, 'y')

        cp = gcrs.cartesian.transform(rz).transform(ry)
        return cp.y.to(u.km).value, cp.z.to(u.km).value

    def sidereal_time(self, time, mode='local'):
        """ Calculates the Apparent Sidereal Time at a reference time

        Parameters:
            time (str,Time): Reference time to calculate sidereal time.
            mode (str): local or greenwich
                If 'local': calculates the sidereal time for the coordinates of this object.
                If 'greenwich': calculates the Greenwich Apparent Sidereal Time.

        Returns:
            sidereal_time: An Astropy Longitude object with the Sidereal Time.
        """
        # return local or greenwich sidereal time
        time = test_attr(time, Time, 'time')
        time.location = self.site
        if mode == 'local':
            return time.sidereal_time('apparent')
        elif mode == 'greenwich':
            return time.sidereal_time('apparent', 'greenwich')
        else:
            raise ValueError('mode must be "local" or "greenwich"')

    def altaz(self, time, coord):
        """ Calculates the Altitude and Azimuth at a reference time for a coordinate

        Parameters:
            time (str,Time): Reference time to calculate the sidereal time.
            coord (str, astropy.SkyCoord): Coordinate of the target ICRS.

        Returns:
            altitude (float): object altitude in degrees.
            azimuth (float): object azimuth in degrees.
        """
        time = test_attr(time, Time, 'time')
        if type(coord) == str:
            coord = SkyCoord(coord, unit=(u.hourangle, u.deg))
        ephem_altaz = coord.transform_to(AltAz(obstime=time, location=self.site))
        return ephem_altaz.alt.deg, ephem_altaz.az.deg

    @property
    def name(self):
        return self.__name

    @property
    def lon(self):
        return self.site.lon

    @lon.setter
    def lon(self, lon):
        lat = self.site.lat
        height = self.site.height
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def lat(self):
        return self.site.lat

    @lat.setter
    def lat(self, lat):
        lon = self.site.lon
        height = self.site.height
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def height(self):
        return self.site.height

    @height.setter
    def height(self, height):
        lon = self.site.lon
        lat = self.site.lat
        site = EarthLocation(lon, lat, height)
        self.site = site

    def __str__(self):
        """ String representation of the Observer class
        """
        out = ('Site: {}\n'
               'Geodetic coordinates: Lon: {}, Lat: {}, height: {:.3f}'.format(
                   self.name, self.site.lon.__str__(), self.site.lat.__str__(),
                   self.site.height.to(u.km))
               )
        return out

    def __del__(self):
        """ When this object is deleted, it removes the name from the Class name list.
        """
        try:
            self.__names.remove(self.__name)
        except:
            pass
